from functools import reduce

from django.db import models
from django.db.models import F, Sum
import six

from . import exceptions


class QuotaLimitField(models.IntegerField):
    """ Django virtual model field.
        Could be used to manage quotas transparently as model fields.
    """

    concrete = False

    def __init__(self, quota_field=None, *args, **kwargs):
        super(QuotaLimitField, self).__init__(*args, **kwargs)
        self._quota_field = quota_field

    def db_type(self, connection):
        # virtual field -- ignore in migrations
        return None

    def contribute_to_class(self, cls, name):
        self.model = cls
        self.name = self.attname = name
        # setting column as none will tell django to not consider this a concrete field
        self.column = None
        # connect myself as the descriptor for this field
        setattr(cls, name, property(self._get_func(), self._set_func()))
        cls._meta.add_field(self, private=True)

    def deconstruct(self, *args, **kwargs):
        name, path, args, kwargs = super(QuotaField, self).deconstruct(*args, **kwargs)
        return (name, path, args, {'default': kwargs.get('default'), 'to': None})

    def _get_func(self):
        # retrieve quota limit from related object
        def func(instance, quota_field=self._quota_field):
            if instance is None:
                raise AttributeError("Can only be accessed via instance")
            try:
                return instance.quotas.get(name=quota_field).limit
            except instance.quotas.model.DoesNotExist:
                return quota_field.default_limit

        return func

    def _set_func(self):
        # store quota limit as related object
        def func(instance, value, quota_field=self._quota_field):
            # a hook to properly init quota after object saved to DB
            quota_field.scope_default_limit(instance, value)
            instance.set_quota_limit(quota_field, value)

        return func


class FieldsContainerMeta(type):
    """ Initiates quota fields names.

        Quotas fields should be located in class with FieldsContainerMeta metaclass.
        Example:
            example_quota = QuotaField()  # this quota field will have name 'example_quota'
    """

    def __new__(self, name, bases, attrs):
        for key in attrs:
            if isinstance(attrs[key], QuotaField):
                attrs[key].name = key
        return type.__new__(self, name, bases, attrs)


class QuotaField(object):
    """ Base quota field.

    Links quota to its scope right after its creation.
    Allows to define:
     - default_limit
     - default_usage
     - is_backend - is quota represents backend limitation. It is impossible to modify backend quotas.
     - creation_condition - function that receive quota scope and return True if quota should be created
                            for given scope. Quota will be created automatically if creation_condition is None.

    Default limit and usage can be defined as callable function.
    Example:
        quota_name = QuotaField(default_limit=lambda scope: scope.attr)
    """

    def __init__(self, default_limit=-1, default_usage=0, is_backend=False, creation_condition=None):
        self.default_limit = default_limit
        self.default_usage = default_usage
        self.is_backend = is_backend
        self.creation_condition = creation_condition

    def is_connected_to_scope(self, scope):
        if self.creation_condition is None:
            return True
        return self.creation_condition(scope)

    def scope_default_limit(self, scope, value=None):
        attr_name = '_default_quota_limit_%s' % self.name
        if value is not None:
            setattr(scope, attr_name, value)
        try:
            return getattr(scope, attr_name)
        except AttributeError:
            return self.default_limit(scope) if six.callable(self.default_limit) else self.default_limit

    def get_or_create_quota(self, scope):
        if not self.is_connected_to_scope(scope):
            raise exceptions.CreationConditionFailedQuotaError(
                'Wrong scope: Cannot create quota "%s" for scope "%s".' % (self.name, scope))
        defaults = {
            'limit': self.scope_default_limit(scope),
            'usage': self.default_usage(scope) if six.callable(self.default_usage) else self.default_usage,
        }

        return scope.quotas.get_or_create(name=self.name, defaults=defaults)

    def get_aggregator_quotas(self, quota):
        """ Fetch ancestors quotas that have the same name and are registered as aggregator quotas. """
        ancestors = quota.scope.get_quota_ancestors()
        aggregator_quotas = []
        for ancestor in ancestors:
            for ancestor_quota_field in ancestor.get_quotas_fields(field_class=AggregatorQuotaField):
                if ancestor_quota_field.get_child_quota_name() == quota.name:
                    aggregator_quotas.append(ancestor.quotas.get(name=ancestor_quota_field))
        return aggregator_quotas

    def __str__(self):
        return self.name

    def recalculate(self, scope):
        if not self.is_connected_to_scope(scope):
            return
        self.recalculate_usage(scope)

    def recalculate_usage(self, scope):
        pass


class CounterQuotaField(QuotaField):
    """ Provides limitation on target models instances count.

    Automatically increases/decreases usage on target instances creation/deletion.

    By default usage is increased by 1. You may tweak this delta by defining get_delta function,
    which accepts target instance and returns number.

    Example:
        # This quota will increase/decrease own values on any resource creation/deletion
        nc_resource_count = CounterQuotaField(
            target_models=lambda: Resource.get_all_models(),  # list or function that return list of target models
            path_to_scope='service_project_link.project',  # path from target model to scope
        )

    It is possible to define trickier calculation by passing `get_current_usage` function as parameter.
    Function should accept two parameters:
        - models - list of target models
        - scope - quota scope
    And return count of current usage.
    """

    def __init__(self, target_models, path_to_scope, get_current_usage=None, get_delta=None, **kwargs):
        self._raw_target_models = target_models
        self._raw_get_current_usage = get_current_usage
        self._raw_get_delta = get_delta
        self.path_to_scope = path_to_scope
        super(CounterQuotaField, self).__init__(**kwargs)

    def get_delta(self, target_instance):
        if not self._raw_get_delta:
            return 1
        return self._raw_get_delta(target_instance)

    def get_current_usage(self, models, scope):
        if self._raw_get_current_usage is not None:
            return self._raw_get_current_usage(models, scope)
        else:
            filter_path_to_scope = self.path_to_scope.replace('.', '__')
            return sum([m.objects.filter(**{filter_path_to_scope: scope}).count() for m in models])

    @property
    def target_models(self):
        if not hasattr(self, '_target_models'):
            self._target_models = (self._raw_target_models() if six.callable(self._raw_target_models)
                                   else self._raw_target_models)
        return self._target_models

    def recalculate_usage(self, scope):
        current_usage = self.get_current_usage(self.target_models, scope)
        scope.set_quota_usage(self.name, current_usage)

    def add_usage(self, target_instance, delta):
        scope = self._get_scope(target_instance)
        delta *= self.get_delta(target_instance)
        if self.is_connected_to_scope(scope):
            scope.add_quota_usage(self.name, delta, validate=True)

    def _get_scope(self, target_instance):
        return reduce(getattr, self.path_to_scope.split('.'), target_instance)


class TotalQuotaField(CounterQuotaField):
    """
    This field aggregates sum of value for the same field of children objects.
    For example, it allows to compute total volume size for the project.

     class Quotas(quotas_models.QuotaModelMixin.Quotas):
        nc_volume_size = quotas_fields.TotalQuotaField(
            target_models=lambda: Volume.get_all_models(),
            path_to_scope='project',
            target_field='size',
        )

    """

    def __init__(self, target_models, path_to_scope, target_field):
        self.target_field = target_field
        super(TotalQuotaField, self).__init__(target_models, path_to_scope)

    def get_current_usage(self, models, scope):
        total_usage = 0
        filter_path_to_scope = self.path_to_scope.replace('.', '__')
        query = {filter_path_to_scope: scope}
        for model in models:
            resources = model.objects.filter(**query)
            subtotal = (resources.values(self.target_field)
                        .aggregate(total_usage=Sum(self.target_field))['total_usage'])
            if subtotal:
                total_usage += subtotal
        return total_usage

    def get_delta(self, target_instance):
        return getattr(target_instance, self.target_field)


class AggregatorQuotaField(QuotaField):
    """ Aggregates sum of quota scope children with the same name.

        Automatically increases/decreases usage if corresponding child quota <aggregation_field> changed.

        Example:
            # This quota will store sum of all customer projects resources
            nc_resource_count = quotas_fields.UsageAggregatorQuotaField(
                get_children=lambda customer: customer.projects.all(),
            )
    """
    aggregation_field = NotImplemented

    def __init__(self, get_children, child_quota_name=None, **kwargs):
        self.get_children = get_children
        self._child_quota_name = child_quota_name
        super(AggregatorQuotaField, self).__init__(**kwargs)

    def get_child_quota_name(self):
        return self._child_quota_name if self._child_quota_name is not None else self.name

    def recalculate_usage(self, scope):
        children = self.get_children(scope)
        current_usage = 0
        for child in children:
            child_quota = child.quotas.get(name=self.get_child_quota_name())
            current_usage += getattr(child_quota, self.aggregation_field)
        scope.set_quota_usage(self.name, current_usage)

    def post_child_quota_save(self, scope, child_quota, created=False):
        current_value = getattr(child_quota, self.aggregation_field)
        if created:
            diff = current_value
        else:
            diff = current_value - child_quota.tracker.previous(self.aggregation_field)
        if diff:
            scope.quotas.filter(name=self.name).update(usage=F('usage') + diff)

    def pre_child_quota_delete(self, scope, child_quota):
        diff = getattr(child_quota, self.aggregation_field)
        if diff:
            scope.quotas.filter(name=self.name).update(usage=F('usage') - diff)


class UsageAggregatorQuotaField(AggregatorQuotaField):
    """ Aggregates sum children quotas usages.

        Note! It is impossible to aggregate usage of another usage aggregator quotas.
        This restriction was added to avoid calls duplications on quota usage field update.
    """
    aggregation_field = 'usage'


class LimitAggregatorQuotaField(AggregatorQuotaField):
    """ Aggregates sum children quotas limits. """
    aggregation_field = 'limit'

# TODO: Implement GlobalQuotaField and GlobalCounterQuotaField



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from django.db import migrations, models
import django.db.models.deletion


def migrate_volume_instances(apps, schema_editor):
    Volume = apps.get_model('openstack', 'Volume')

    for volume in Volume.objects.all():
        instance = volume.instances.all().first()
        if instance:
            volume.instance = instance
            volume.save()


class Migration(migrations.Migration):

    dependencies = [
        ('openstack', '0020_tenant_extra_configuration'),
    ]

    operations = [
        migrations.AddField(
            model_name='volume',
            name='instance',
            field=models.ForeignKey(related_name='+', blank=True, to='openstack.Instance', null=True, on_delete=django.db.models.deletion.SET_NULL),
        ),
        migrations.RunPython(migrate_volume_instances, elidable=True),
        migrations.RemoveField(
            model_name='instance',
            name='volumes',
        ),
        migrations.AlterField(
            model_name='volume',
            name='instance',
            field=models.ForeignKey(related_name='volumes', blank=True, to='openstack.Instance', null=True, on_delete=django.db.models.deletion.SET_NULL),
        ),
    ]



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from django.db import models, migrations


def pull_security_group_tenant(apps, schema_editor):
    SecurityGroup = apps.get_model('openstack', 'SecurityGroup')
    for group in SecurityGroup.objects.all():
        group.tenant = group.service_project_link.tenants.first()
        group.save()


class Migration(migrations.Migration):

    dependencies = [
        ('openstack', '0008_require_instance_tenant'),
    ]

    operations = [
        migrations.AddField(
            model_name='securitygroup',
            name='tenant',
            field=models.ForeignKey(related_name='security_groups', to='openstack.Tenant', null=True),
        ),
        migrations.RunPython(pull_security_group_tenant, elidable=True)
    ]



from django.utils.functional import cached_property

from waldur_core.structure import ServiceBackend

from .client import RancherClient
from . import models


class RancherBackend(ServiceBackend):
    def __init__(self, settings):
        """
        :type settings: :class:`waldur_core.structure.models.ServiceSettings`
        """
        self.settings = settings

    @cached_property
    def client(self):
        """
        Construct Rancher REST API client using credentials specified in the service settings.
        """
        client = RancherClient(self.host, verify_ssl=False)
        client.login(self.settings.username, self.settings.password)
        return client

    @cached_property
    def host(self):
        return self.settings.backend_url.strip('/')

    def create_cluster(self, cluster):
        backend_cluster = self.client.create_cluster(cluster.name)
        self._backend_cluster_to_cluster(backend_cluster, cluster)
        self.client.create_cluster_registration_token(cluster.backend_id)
        cluster.node_command = self.client.get_node_command(cluster.backend_id)
        cluster.save()

    def delete_cluster(self, cluster):
        self.client.delete_cluster(cluster.backend_id)

    def update_cluster(self, cluster):
        backend_cluster = self._cluster_to_backend_cluster(cluster)
        self.client.update_cluster(cluster.backend_id, backend_cluster)

    def _backend_cluster_to_cluster(self, backend_cluster, cluster):
        cluster.backend_id = backend_cluster['id']
        cluster.name = backend_cluster['name']

    def _cluster_to_backend_cluster(self, cluster):
        return {'name': cluster.name}

    def get_clusters_for_import(self):
        cur_clusters = set(models.Cluster.objects.filter(service_project_link__service__settings=self.settings)
                           .values_list('backend_id', flat=True))
        backend_clusters = self.client.list_clusters()
        return filter(lambda c: c['id'] not in cur_clusters, backend_clusters)

    def import_cluster(self, backend_id, service_project_link):
        backend_cluster = self.client.get_cluster(backend_id)
        cluster = models.Cluster(
            backend_id=backend_id,
            service_project_link=service_project_link,
            state=models.Cluster.States.OK)
        self._backend_cluster_to_cluster(backend_cluster, cluster)
        cluster.save()
        backend_nodes = backend_cluster.get('appliedSpec', {}).get('rancherKubernetesEngineConfig', {}).get('nodes', [])
        for node in backend_nodes:
            models.Node.objects.create(
                cluster=cluster
            )
        return cluster



# -*- coding: utf-8 -*-
# Generated by Django 1.11.13 on 2018-10-04 10:41
from __future__ import unicode_literals

from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    dependencies = [
        ('marketplace', '0039_split_component_model'),
    ]

    operations = [
        migrations.CreateModel(
            name='CategoryColumn',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('index', models.PositiveSmallIntegerField(help_text='Index allows to reorder columns.')),
                ('title', models.CharField(help_text='Title is rendered as column header.', max_length=255)),
                ('attribute', models.CharField(blank=True, help_text='Resource attribute is rendered as table cell.', max_length=255)),
                ('widget', models.CharField(blank=True, help_text='Widget field allows to customise table cell rendering.', max_length=255)),
                ('category', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='columns', to='marketplace.Category')),
            ],
            options={
                'ordering': ('category', 'index'),
            },
        ),
    ]



import pickle  # nosec

import six

from . import ServiceBackendError


class SerializableBackendError(ServiceBackendError):
    def __init__(self, *args, **kwargs):
        if not args:
            super(SerializableBackendError, self).__init__(*args, **kwargs)

        # Some client exceptions, such as cinder_exceptions.ClientException
        # are not serializable by Celery, because they use custom arguments *args
        # and define __init__ method, but don't call Exception.__init__ method
        # http://docs.celeryproject.org/en/latest/userguide/tasks.html#creating-pickleable-exceptions
        # That's why when Celery worker tries to deserialize OpenStack client exception,
        # it uses empty invalid *args. It leads to unrecoverable error and worker dies.
        # When all workers are dead, all tasks are stuck in pending state forever.
        # In order to fix this issue we serialize exception to text type explicitly.
        args = list(args)
        for i, arg in enumerate(args):
            try:
                # pickle is used to check celery internal errors serialization,
                # it is safe from security point of view
                pickle.loads(pickle.dumps(arg))  # nosec
            except (pickle.PickleError, TypeError):
                args[i] = six.text_type(arg)

        super(SerializableBackendError, self).__init__(*args, **kwargs)



# -*- coding: utf-8 -*-
# Generated by Django 1.11.7 on 2018-06-06 12:36
from __future__ import unicode_literals

from django.db import migrations, models


def updated_offerings(apps, schema_editor):
    # Update value of terminated_at field with value of modified field
    Offering = apps.get_model('support', 'Offering')
    Offering.objects.filter(state='terminated').update(terminated_at=models.F('modified'))


class Migration(migrations.Migration):

    dependencies = [
        ('support', '0017_unique_backend_id'),
    ]

    operations = [
        migrations.AddField(
            model_name='offering',
            name='terminated_at',
            field=models.DateTimeField(blank=True, editable=False, null=True),
        ),
        migrations.RunPython(updated_offerings, elidable=True)
    ]



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from django.db import migrations, models
import django.utils.timezone
import model_utils.fields
import waldur_core.core.fields


class Migration(migrations.Migration):

    dependencies = [
        ('cost_tracking', '0022_priceestimate_leafs'),
    ]

    operations = [
        migrations.CreateModel(
            name='ConsumptionDetails',
            fields=[
                ('id', models.AutoField(verbose_name='ID', serialize=False, auto_created=True, primary_key=True)),
                ('created', model_utils.fields.AutoCreatedField(default=django.utils.timezone.now, verbose_name='created', editable=False)),
                ('modified', model_utils.fields.AutoLastModifiedField(default=django.utils.timezone.now, verbose_name='modified', editable=False)),
                ('uuid', waldur_core.core.fields.UUIDField()),
                ('configuration', waldur_core.core.fields.JSONField(default={}, help_text='Current resource configuration.')),
                ('last_update_time', models.DateTimeField(help_text='Last configuration change time.')),
                ('consumed_before_update', waldur_core.core.fields.JSONField(default={}, help_text='How many consumables were used by resource before last update.')),
                ('price_estimate', models.OneToOneField(related_name='consumption_details', to='cost_tracking.PriceEstimate')),
            ],
            options={
                'abstract': False,
                'verbose_name': 'Consumption details',
                'verbose_name_plural': 'Consumption details',
            },
        ),
    ]



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

import os

from django.core.management.base import BaseCommand
from django.core.files import File

from waldur_mastermind.marketplace.models import Category, CategoryColumn, \
    CategoryComponent, Section, Attribute, AttributeOption


def merge_two_dicts(x, y):
    z = x.copy()   # start with x's keys and values
    z.update(y)    # modifies z with y's keys and values & returns None
    return z


def humanize(name):
    return name.replace("_", " ").capitalize()


available_categories = {
    'backup': ('Backup', 'Backup solution'),
    'consultancy': ('Consultancy', 'Experts for hire'),
    'collocation': ('Collocation', 'Collocation services'),
    'cms': ('CMS', 'Content Management Systems'),
    'db': ('Databases', 'Relational DBMS'),
    'email': ('E-mail', 'E-mail services'),
    'hpc': ('HPC', 'High Performance Computing'),
    'licenses': ('Licenses', 'Application and OS licenses'),
    'vm': ('VMs', 'Virtual machines'),
    'vpc': ('Private clouds', 'Virtual private clouds'),
    'network': ('Network', 'Network services'),
    'operations': ('Operations', 'Reliable support'),
    'security': ('Security', 'Security services'),
    'storage': ('Storage', 'Data preservation'),
    # devices
    'measurement_devices': ('Measurement systems', 'Available measurement devices'),
}

category_columns = {
    'storage': [
        {
            'title': 'Size',
            'widget': 'filesize',
            'attribute': 'size',
        },
        {
            'title': 'Attached to',
            'widget': 'attached_instance',
        },
    ],
    'vm': [
        {
            'title': 'Internal IP',
            'attribute': 'internal_ips',
            'widget': 'csv',
        },
        {
            'title': 'External IP',
            'attribute': 'external_ips',
            'widget': 'csv',
        },
    ],
}

category_components = {
    'vpc': [
        {
            'type': 'storage',
            'name': 'Storage',
            'measured_unit': 'GB',
        },
        {
            'type': 'ram',
            'name': 'RAM',
            'measured_unit': 'GB',
        },
        {
            'type': 'cores',
            'name': 'Cores',
            'measured_unit': 'cores',
        }
    ]
}

common_sections = {
    'Support': [
        ('email', 'E-mail', 'string'),
        ('phone', 'Phone', 'string'),
        ('portal', 'Support portal', 'string'),
        ('description', 'Description', 'string'),
        ('terms_of_services_link', 'ToS link', 'string'),
    ],
    'SLA_simple': [
        ('low_sla_response', 'Response time (low priority, mins)', 'integer'),
        ('medium_sla_response', 'Response time (medium priority, mins)', 'integer'),
        ('high_sla_response', 'Response time (high priority, mins)', 'integer'),
    ],
    'SLA': [
        ('low_sla_response_wh', 'Response time (low priority, working hours)', 'integer'),
        ('low_sla_resolution_wh', 'Resolution time (low priority, working hours)', 'integer'),
        ('low_sla_response_nwh', 'Response time (low priority, non-working hours)', 'integer'),
        ('low_sla_resolution_nwh', 'Resolution time (low priority, non-working hours)', 'integer'),

        ('medium_sla_response_wh', 'Response time (medium priority, working hours)', 'integer'),
        ('medium_sla_resolution_wh', 'Resolution time (medium priority, working hours)', 'integer'),
        ('medium_sla_response_nwh', 'Response time (medium priority, non-working hours)', 'integer'),
        ('medium_sla_resolution_nwh', 'Resolution time (medium priority, non-working hours)', 'integer'),

        ('high_sla_response_wh', 'Response time (high priority, working hours)', 'integer'),
        ('high_sla_resolution_wh', 'Resolution time (high priority, working hours)', 'integer'),
        ('high_sla_response_nwh', 'Response time (high priority, non-working hours)', 'integer'),
        ('high_sla_resolution_nwh', 'Resolution time (high priority, non-working hours)', 'integer'),
    ],
    'Security': [
        ('certification', 'Certification', 'list'),
    ],

    'Location': [
        ('address', 'Address', 'string')
    ]
}

hpc_sections = {
    'system_information': [
        ('queuing_system', 'Queueing system', 'list'),
        ('home_space', 'Home space', 'string'),
        ('work_space', 'Work space', 'string'),
        ('linux_distro', 'Linux distribution', 'list'),
    ],
    'node_information': [
        ('cpu', 'CPU model', 'choice'),
        ('gpu', 'GPU model', 'choice'),
        ('memory', 'Memory per node (GB)', 'integer'),
        ('local_disk', 'Local disk (GB)', 'integer'),
        ('interconnect', 'Interconnect', 'choice'),
        ('node_count', 'Node count', 'integer'),
    ],
    'performance': [
        ('tflops', 'Peak TFlop/s', 'integer'),
        ('linpack', 'Linpack TFlop/s', 'integer')
    ],
    'software': [
        ('applications', 'Applications', 'list'),
    ],
}

collocation_sections = {
    'features': [
        ('collocation_remote_access', 'Remote access', 'choice'),
        ('computing_network', 'Network access', 'list'),
        ('collocation_dimensions', 'Dimensions', 'list'),
        ('collocation_power', 'Power (A)', 'integer'),
        ('collocation_cages', 'Locked cages', 'boolean'),
    ],
}

measurement_devices_sections = {
    'properties': [
        ('measurement_devices_type', 'Type', 'choice'),
        ('measurement_devices_sample_matrix', 'Sample matrix', 'list'),
        ('measurement_devices_principle', 'Measurement principle', 'list'),
    ],
    'model': [
        ('measurement_devices_mark', 'Mark', 'string'),
        ('measurement_devices_model', 'Model', 'string'),
        ('measurement_devices_manufacturer', 'Manufacturer', 'string')
    ],
    'usage_info': [
        ('measurement_devices_routine_application', 'Routine applications', 'text'),
        ('measurement_devices_additional_components', 'Additional components', 'list')
    ]
}

computing_common_sections = {
    'details': [
        ('virtualization', 'Virtualization', 'choice'),
        ('computing_network', 'Network', 'list'),
        ('ha', 'High Availability', 'boolean'),
        ('av_monitoring', 'Availability monitoring', 'boolean'),
    ],
    'application': [
        ('os', 'Operating system', 'list'),
        ('application', 'Application', 'list'),
    ],
}

vpc_sections = {
    # nothing yet, TBA
}

vm_sections = {
    'software': [
        ('antivirus', 'Antivirus', 'boolean'),
    ],
    'remote_access': [
        ('vm_remote_access', 'Remote access', 'list'),
        ('vm_access_level', 'Access level', 'choice'),
    ]
}

email_sections = {
    'software': [
        ('email_software', 'Software', 'choice'),
    ],
    'features': [
        ('delegated_domain_administration', 'Delegated domain administration', 'boolean'),
        ('calendar', 'Calendar management', 'boolean'),
        ('webchat', 'Webchat', 'boolean'),
    ]
}

storage_sections = {
    'details': [
        ('storage_type', 'Storage type', 'choice'),
    ],
    'access': [
        ('web_interface', 'Web interface', 'boolean'),
        ('api', 'API', 'boolean'),
        ('api_flavor', 'API flavor', 'list'),
    ],
    'encryption': [
        ('encryption_at_rest', 'Encryption at-rest', 'boolean'),
        ('encryption_in_transit', 'Encryption in-transit', 'boolean'),
    ]
}

common_expert_sections = {
    'Scope': [
        ('scope_of_services', 'Scope of services', 'list'),
    ]
}

operations_sections = {
    'Supported services': [
        ('os', 'Supported OS', 'list'),
        ('application', 'Supported applications', 'list'),
    ],
}

consultancy_sections = {
    # nothing yet, TBA
}

security_sections = {
    'Application': [
        ('security_application', 'Application', 'string'),
        ('hardware_module', 'Hardware module', 'boolean'),
        ('vendor_name', 'Vendor name', 'string'),
        ('application_version', 'Application version', 'string'),
    ],
    'Access': [
        ('security_access', 'Access', 'list'),
    ]
}

network_section = {
    'Technology': [
        ('computing_network', 'Connected network', 'list'),
        ('vpn_technology', 'VPN technology', 'list'),
    ]
}

specific_sections = {
    'collocation': collocation_sections,
    'computing': merge_two_dicts(computing_common_sections, vm_sections),
    'consultancy': merge_two_dicts(common_expert_sections, consultancy_sections),
    'email': email_sections,
    'hpc': hpc_sections,
    'operations': merge_two_dicts(common_expert_sections, operations_sections),
    'vm': merge_two_dicts(computing_common_sections, vm_sections),
    'vpc': merge_two_dicts(computing_common_sections, vpc_sections),
    'security': security_sections,
    'measurement_devices': measurement_devices_sections,
    'storage': storage_sections,
}

enums = {
    'languages': [
        ('et', 'Estonian'),
        ('en', 'English'),
        ('lv', 'Latvian'),
        ('lt', 'Lithuanian'),
        ('ru', 'Russian'),
        ('sw', 'Swedish'),
        ('fi', 'Finnish'),
    ],
    'collocation_dimensions': [
        ('600x800', '600 x 800'),
        ('600x1000', '600 x 1000'),
    ],
    'deployment_type': [
        ('appliance', 'Appliance (Managed)'),
        ('remote', 'Remote (SaaS)')
    ],
    'email_software': [
        ('zimbra', 'Zimbra'),
        ('ibm_lotus', 'IBM Lotus'),
    ],
    'workdays': [
        ('base', '5 days'),
        ('extended', '7 days'),
    ],
    'businesshours': [
        ('basehours', '8 hours'),
        ('extendedhours', '24 hours'),
    ],
    'priority': [
        ('eob', 'End-of-business day'),
        ('nbd', 'Next business day'),
    ],
    'certification': [
        ('iskel', 'ISKE L'),
        ('iskem', 'ISKE M'),
        ('iskeh', 'ISKE H'),
        ('iso27001', 'ISO27001'),
    ],
    'interconnect': [
        ('infiniband_fdr', 'Infiniband FDR'),
        ('infiniband_edr', 'Infiniband EDR'),
        ('Ethernet_1G', 'Ethernet 1G'),
        ('Ethernet_10G', 'Ethernet 10G'),
    ],
    'sla_response': [
        '1', '1 hour',
        '2', '2 hours',
        '3', '3 hours',
        'eob', 'End of business day',
    ],
    'virtualization': [
        ('kvm', 'KVM'),
        ('xen', 'XEN'),
        ('vmware', 'VMware'),
        ('Baremetal', 'Baremetal'),
    ],
    'measurement_devices_type': [
        ('aas', 'Atomic Absorption Spectrometer'),
        ('spectrophotometer', 'Spectrophotometer'),
        ('spectrometers', 'Spectrometers'),
    ],
    'measurement_devices_sample_matrix': [
        ('liquids', 'Liquids'),
        ('solds', 'Solids'),
    ],
    'measurement_devices_principle': [
        ('x-ray-diffraction', 'X-ray diffraction'),
        ('chromatography', 'Chromatography'),
        ('mass-spectrometry ', 'Mass-spectrometry'),
        ('electrophoresis', 'Electrophoresis'),
        ('magnetic-resonance', 'Magnetic resonance'),
        ('physical-properties', 'Physical properties'),
        ('infrapuna-spektromeeter', 'Infrapuna spektromeeter'),
        ('thz-radiation-amplitude-and-phase', 'Measurement of THz radiation amplitude and phase'),
    ],
    'measurement_devices_additional_components': [
        ('autosampler-with-cooling', 'Autosampler with cooling'),
        ('autosampler', 'Autosampler'),
        ('photodiode-array-detector', 'Photodiode array detector (PDA)'),
        ('column-thermostate', 'Column thermostate'),
        ('flame-ionization-detector', 'Flame ionization detector (FID)'),
        ('thermal-conductivity-detector', 'Thermal conductivity detector (TCD)')
    ],
    'scope_of_services': [
        ('analysis', 'Analysis'),
        ('implementation', 'Implementation'),
        ('design', 'Design'),
        ('deployment', 'Deployment'),
        ('issue_resolution', 'issue_resolution'),
        ('change_management', 'Change management'),
        ('disaster_recovery', 'Disaster recovery'),
    ],
    'security_access': [
        ('api', 'API'),
        ('offline', 'Offline'),
    ],
    'storage_type': [
        ('block', 'Block'),
        ('object', 'Object'),
        ('fs', 'Filesystem'),
    ],
    'computing_network': [
        ('private', 'Private (own)'),
        ('aso', 'ASO'),
        ('ddn', 'DDN'),
        ('ogn', 'OGN'),
        ('banglagovnet', 'BanglaGovNet'),
        ('public', 'Public Internet'),
    ],
    'vpn_technology': [
        ('ipsec', 'IPSEC'),
        ('gre+ipsec', 'GRE + IPSEC'),
    ],
    'vm_access_level': [
        ('root', 'Root / Administrator'),
        ('user', 'User level'),
    ],
    'api_flavor': [
        ('s3', 'S3'),
        ('swift', 'Swift'),
        ('custom', 'Custom'),
    ],
    'vm_remote_access': [
        ('console', 'Console'),
        ('ssh_rdp', 'SSH/RDP'),
        ('direct_access', 'Direct access'),
    ],
    'collocation_remote_access': [
        ('vpn', 'VPN'),
    ],
    'os': [
        ('ubuntu16.04', 'Ubuntu 16.04'),
        ('centos7', 'CentOS 7'),
        ('windows2016', 'Windows 2016'),
        ('rhel7', 'RHEL 7'),
    ],
    'application': [
        ('zevenet', 'Zevenet'),
        ('owncloud', 'Owncloud'),
        ('lamp', 'LAMP'),
        ('nginx', 'Nginx'),
        ('iis', 'IIS'),
        ('mssql2017', 'MS SQL Server 2017'),
        ('mysql57', 'MySQL 5.7'),
    ],
}


def populate_category(category_code, category, sections):
    for section_key in sections.keys():
        section_prefix = '%s_%s' % (category_code, section_key)
        sec, _ = Section.objects.get_or_create(key=section_prefix, title=humanize(section_key), category=category)
        sec.is_standalone = True
        sec.save()
        for attribute in sections[section_key]:
            key, title, attribute_type = attribute
            attr, _ = Attribute.objects.get_or_create(key='%s_%s' % (section_prefix, key),
                                                      title=title, type=attribute_type, section=sec)
            if key in enums:
                values = enums[key]
                for val_key, val_label in values:
                    AttributeOption.objects.get_or_create(
                        attribute=attr, key='%s_%s_%s' % (section_prefix, key, val_key), title=val_label)


def load_category(category_short):
    category_name, category_description = available_categories[category_short]
    new_category, _ = Category.objects.get_or_create(title=category_name, description=category_description)
    category_icon = '%s.svg' % category_short
    path = os.path.join(os.path.dirname(os.path.realpath(__file__)), 'category_icons/')
    new_category.icon.save(category_icon,
                           File(open(path + category_icon, 'r')))
    new_category.save()
    # populate category with common section
    populate_category(category_short, new_category, common_sections)
    # add specific sections
    if category_short in specific_sections.keys():
        populate_category(category_short, new_category, specific_sections[category_short])

    # add category columns
    columns = category_columns.get(category_short, [])
    for index, attribute in enumerate(columns):
        CategoryColumn.objects.get_or_create(
            category=new_category,
            title=attribute['title'],
            defaults=dict(
                index=index,
                attribute=attribute.get('attribute', ''),
                widget=attribute.get('widget'),
            )
        )

    # add category components
    components = category_components.get(category_short, [])
    for component in components:
        CategoryComponent.objects.get_or_create(
            category=new_category,
            type=component['type'],
            defaults=dict(
                name=component['name'],
                measured_unit=component.get('measured_unit'),
            )
        )
    return new_category


class Command(BaseCommand):
    help = 'Loads a categories for the Marketplace'

    missing_args_message = 'Please define at least one category to load, available are:\n%s' %\
                           '\n'.join(available_categories.keys())

    def add_arguments(self, parser):
        parser.add_argument('category', nargs='+', type=str, help='List of categories to load')
        parser.add_argument('--basic_sla', nargs='?', type=bool, default=False, help='Use basic SLA for categories')

    def handle(self, *args, **options):

        all_categories = available_categories.keys()
        if options['basic_sla']:
            del common_sections['SLA']
        for category_short in options['category']:
            if category_short not in all_categories:
                self.stdout.write(self.style.WARNING('Category "%s" is not available' % category_short))
                continue
            new_category = load_category(category_short)
            self.stdout.write(self.style.SUCCESS('Loaded category %s, %s ' % (category_short, new_category.uuid)))



# -*- coding: utf-8 -*-
# Generated by Django 1.9.13 on 2017-04-10 16:46
from __future__ import unicode_literals

from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ('openstack', '0033_remove_openstackservice_name'),
    ]

    operations = [
        migrations.RemoveField(
            model_name='floatingip',
            name='start_time',
        ),
        migrations.RemoveField(
            model_name='network',
            name='start_time',
        ),
        migrations.RemoveField(
            model_name='securitygroup',
            name='start_time',
        ),
        migrations.RemoveField(
            model_name='subnet',
            name='start_time',
        ),
        migrations.RemoveField(
            model_name='tenant',
            name='start_time',
        ),
    ]



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from django.db import migrations, models
import django.db.models.deletion
import django.utils.timezone
import model_utils.fields
import taggit.managers
import django_fsm


class Migration(migrations.Migration):

    dependencies = [
        ('taggit', '0002_auto_20150616_2121'),
        ('openstack_tenant', '0011_backupschedule_call_count'),
    ]

    operations = [
        migrations.AddField(
            model_name='backupschedule',
            name='backend_id',
            field=models.CharField(max_length=255, blank=True),
        ),
        migrations.AddField(
            model_name='backupschedule',
            name='created',
            field=model_utils.fields.AutoCreatedField(default=django.utils.timezone.now, verbose_name='created', editable=False),
        ),
        migrations.AddField(
            model_name='backupschedule',
            name='modified',
            field=model_utils.fields.AutoLastModifiedField(default=django.utils.timezone.now, verbose_name='modified', editable=False),
        ),
        migrations.AddField(
            model_name='backupschedule',
            name='service_project_link',
            field=models.ForeignKey(related_name='backup_schedules', on_delete=django.db.models.deletion.PROTECT, to='openstack_tenant.OpenStackTenantServiceProjectLink', null=True),
            preserve_default=False,
        ),
        migrations.AddField(
            model_name='backupschedule',
            name='start_time',
            field=models.DateTimeField(null=True, blank=True),
        ),
        migrations.AddField(
            model_name='backupschedule',
            name='state',
            field=django_fsm.FSMIntegerField(default=5, choices=[(5, 'Creation Scheduled'), (6, 'Creating'), (1, 'Update Scheduled'), (2, 'Updating'), (7, 'Deletion Scheduled'), (8, 'Deleting'), (3, 'OK'), (4, 'Erred')]),
        ),
        migrations.AddField(
            model_name='backupschedule',
            name='tags',
            field=taggit.managers.TaggableManager(to='taggit.Tag', through='taggit.TaggedItem', blank=True, help_text='A comma-separated list of tags.', verbose_name='Tags'),
        ),
    ]



# -*- coding: utf-8 -*-
# Generated by Django 1.11.13 on 2018-09-17 18:15
from __future__ import unicode_literals

from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    dependencies = [
        ('marketplace', '0035_component_name_unit_description'),
    ]

    operations = [
        migrations.AddField(
            model_name='serviceprovider',
            name='api_secret_code',
            field=models.CharField(blank=True, max_length=255, null=True),
        ),
        migrations.AlterField(
            model_name='serviceprovider',
            name='customer',
            field=models.OneToOneField(on_delete=django.db.models.deletion.CASCADE, to='structure.Customer'),
        ),
    ]



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from django.db import migrations


def remove_duplicate_floating_ips(apps, schema_editor):
    already_seen = set()
    FloatingIP = apps.get_model('openstack_tenant', 'FloatingIP')

    for fip in FloatingIP.objects.iterator():
        key = (fip.settings_id, fip.backend_id)
        if key in already_seen:
            fip.delete()
        else:
            already_seen.add(key)


class Migration(migrations.Migration):

    dependencies = [
        ('openstack_tenant', '0026_remove_start_time'),
    ]

    operations = [
        migrations.RunPython(remove_duplicate_floating_ips, elidable=True),
    ]



# -*- coding: utf-8 -*-
# Generated by Django 1.11.13 on 2018-10-12 09:33
from __future__ import unicode_literals

from django.db import migrations, models
import django.db.models.deletion
import django.utils.timezone
import waldur_mastermind.invoices.models


class Migration(migrations.Migration):

    dependencies = [
        ('packages', '0014_unique_package_tenant'),
        ('invoices', '0024_immutable_default_json'),
    ]

    operations = [
        migrations.CreateModel(
            name='ServiceDowntime',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('start', models.DateTimeField(default=waldur_mastermind.invoices.models.get_default_downtime_start, help_text='Date and time when downtime has started.')),
                ('end', models.DateTimeField(default=django.utils.timezone.now, help_text='Date and time when downtime has ended.')),
                ('package', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='packages.OpenStackPackage')),
            ],
        ),
    ]



from django.db import transaction

from waldur_core.logging import tasks


def process_hook(sender, instance, created=False, **kwargs):
    transaction.on_commit(lambda: tasks.process_event.delay(instance.pk))



from . import views


def register_in(router):
    router.register(r'rancher', views.ServiceViewSet, base_name='rancher')
    router.register(r'rancher-spl', views.ServiceProjectLinkViewSet,
                    base_name='rancher-spl')
    router.register(r'rancher-clusters', views.ClusterViewSet,
                    base_name='rancher-cluster')
    router.register(r'rancher-nodes', views.NodeViewSet,
                    base_name='rancher-node')



"""
TODO: drop patch when django-fsm package is updated.

If model with FSM state field has other fields that access their field value
via a property or a virtual Field, then creation of instances will fail.

There is pending patch in upstream project:
https://github.com/kmmbvnr/django-fsm/pull/171
"""

__all__ = ['monkey_patch_fields']


def subfield_get(self, obj, type=None):
    """
    Verbatim copy from:
    https://github.com/django/django/blob/1.9.13/django/db/models/fields/subclassing.py#L38
    """
    if obj is None:
        return self
    return obj.__dict__[self.field.name]


def get_field_name(self):
    return self.field.name


def patch_field_descriptor(cls):
    cls.__get__ = subfield_get
    setattr(cls, 'field_name', property(get_field_name))


def patch_fsm_field_mixin(cls):
    from django_fsm import TransitionNotAllowed, pre_transition, post_transition

    def change_state(self, instance, method, *args, **kwargs):
        meta = method._django_fsm
        method_name = method.__name__
        current_state = self.get_state(instance)
        try:
            current_state_name = filter(lambda x: x[0] == current_state, meta.field.choices)[0][1]
        except Exception:
            current_state_name = current_state

        if not meta.has_transition(current_state):
            raise TransitionNotAllowed(
                "Can't switch from state '{0}' using method '{1}'".format(current_state_name, method_name),
                object=instance, method=method)
        if not meta.conditions_met(instance, current_state):
            raise TransitionNotAllowed(
                "Transition conditions have not been met for method '{0}'".format(method_name),
                object=instance, method=method)

        next_state = meta.next_state(current_state)

        signal_kwargs = {
            'sender': instance.__class__,
            'instance': instance,
            'name': method_name,
            'source': current_state,
            'target': next_state
        }

        pre_transition.send(**signal_kwargs)

        try:
            result = method(instance, *args, **kwargs)
            if next_state is not None:
                self.set_proxy(instance, next_state)
                self.set_state(instance, next_state)
        except Exception as exc:
            exception_state = meta.exception_state(current_state)
            if exception_state:
                self.set_proxy(instance, exception_state)
                self.set_state(instance, exception_state)
                signal_kwargs['target'] = exception_state
                signal_kwargs['exception'] = exc
                post_transition.send(**signal_kwargs)
            raise
        else:
            post_transition.send(**signal_kwargs)

        return result

    cls.change_state = change_state


def monkey_patch_fields():
    from django_fsm import FSMFieldDescriptor, FSMFieldMixin
    patch_field_descriptor(FSMFieldDescriptor)
    patch_fsm_field_mixin(FSMFieldMixin)



from __future__ import unicode_literals

import base64
import os

import pdfkit
from PIL import Image
from django.conf import settings
from django.core.files.base import ContentFile
from django.core.files.storage import default_storage as storage
from django.template.loader import render_to_string
from django.utils import six, timezone
from django.utils.translation import ugettext_lazy as _
from rest_framework import exceptions, serializers

from waldur_core.core import utils as core_utils

from . import models, plugins


def get_order_item_processor(order_item):
    if order_item.resource:
        offering = order_item.resource.offering
    else:
        offering = order_item.offering

    if order_item.type == models.RequestTypeMixin.Types.CREATE:
        return plugins.manager.get_processor(offering.type, 'create_resource_processor')

    elif order_item.type == models.RequestTypeMixin.Types.UPDATE:
        return plugins.manager.get_processor(offering.type, 'update_resource_processor')

    elif order_item.type == models.RequestTypeMixin.Types.TERMINATE:
        return plugins.manager.get_processor(offering.type, 'delete_resource_processor')


def process_order_item(order_item, user):
    processor = get_order_item_processor(order_item)
    if not processor:
        order_item.error_message = 'Skipping order item processing because processor is not found.'
        order_item.set_state_erred()
        order_item.save(update_fields=['state', 'error_message'])
        return

    try:
        order_item.set_state_executing()
        order_item.save(update_fields=['state'])
        processor(order_item).process_order_item(user)
    except exceptions.APIException as e:
        order_item.error_message = six.text_type(e)
        order_item.set_state_erred()
        order_item.save(update_fields=['state', 'error_message'])


def validate_order_item(order_item, request):
    processor = get_order_item_processor(order_item)
    if processor:
        try:
            processor(order_item).validate_order_item(request)
        except NotImplementedError:
            # It is okay if validation is not implemented yet
            pass


def create_screenshot_thumbnail(screenshot):
    pic = screenshot.image
    fh = storage.open(pic.name, 'r')
    image = Image.open(fh)
    image.thumbnail(settings.WALDUR_MARKETPLACE['THUMBNAIL_SIZE'], Image.ANTIALIAS)
    fh.close()

    thumb_extension = os.path.splitext(pic.name)[1]
    thumb_extension = thumb_extension.lower()
    thumb_name = os.path.basename(pic.name)

    if thumb_extension in ['.jpg', '.jpeg']:
        FTYPE = 'JPEG'
    elif thumb_extension == '.gif':
        FTYPE = 'GIF'
    elif thumb_extension == '.png':
        FTYPE = 'PNG'
    else:
        return

    temp_thumb = six.StringIO()
    image.save(temp_thumb, FTYPE)
    temp_thumb.seek(0)
    screenshot.thumbnail.save(thumb_name, ContentFile(temp_thumb.read()), save=True)
    temp_thumb.close()


def create_order_pdf(order):
    logo_path = settings.WALDUR_CORE['SITE_LOGO']
    if logo_path:
        with open(logo_path, 'r') as image_file:
            deployment_logo = base64.b64encode(image_file.read())
    else:
        deployment_logo = None

    context = dict(
        order=order,
        currency=settings.WALDUR_CORE['CURRENCY_NAME'],
        deployment_name=settings.WALDUR_CORE['SITE_NAME'],
        deployment_address=settings.WALDUR_CORE['SITE_ADDRESS'],
        deployment_email=settings.WALDUR_CORE['SITE_EMAIL'],
        deployment_phone=settings.WALDUR_CORE['SITE_PHONE'],
        deployment_logo=deployment_logo,
    )
    html = render_to_string('marketplace/order.html', context)
    pdf = pdfkit.from_string(html, False)
    order.file = base64.b64encode(pdf)
    order.save()


def import_resource_metadata(resource):
    instance = resource.scope
    fields = {'action', 'action_details', 'state', 'runtime_state'}

    for field in fields:
        if field == 'state':
            value = instance.get_state_display()
        else:
            value = getattr(instance, field, None)
        if field in fields:
            resource.backend_metadata[field] = value

    resource.name = instance.name
    resource.save(update_fields=['backend_metadata', 'attributes', 'name'])


def get_service_provider_info(source):
    try:
        resource = models.Resource.objects.get(scope=source)
        customer = resource.offering.customer
        service_provider = getattr(customer, 'serviceprovider', None)

        return {
            'service_provider_name': customer.name,
            'service_provider_uuid': '' if not service_provider else service_provider.uuid.hex,
        }
    except models.Resource.DoesNotExist:
        return {}


def format_list(resources):
    """
    Format comma-separated list of IDs from Django queryset.
    """
    return ', '.join(map(str, sorted(resources.values_list('id', flat=True))))


def get_order_item_url(order_item):
    link_template = settings.WALDUR_MARKETPLACE['ORDER_ITEM_LINK_TEMPLATE']
    return link_template.format(order_item_uuid=order_item.uuid,
                                project_uuid=order_item.order.project.uuid)


def fill_activated_field(apps, schema_editor):
    # We cannot use RequestTypeMixin.Types.CREATE and OrderItem.States.Done because this function called in migrations
    state_done = 3
    type_create = 1

    OrderItem = apps.get_model('marketplace', 'OrderItem')

    for order_item in OrderItem.objects.filter(type=type_create, state=state_done):
        if not order_item.activated and order_item.resource:
            order_item.activated = order_item.resource.created
            order_item.save()


def get_info_about_missing_usage_reports():
    now = timezone.now()
    billing_period = core_utils.month_start(now)

    offering_ids = models.OfferingComponent.objects.filter(billing_type=models.OfferingComponent.BillingTypes.USAGE).\
        values_list('offering_id', flat=True)
    resource_with_usages = models.ComponentUsage.objects.filter(billing_period=billing_period).\
        values_list('resource', flat=True)
    resources_without_usages = models.Resource.objects.\
        filter(state=models.Resource.States.OK, offering_id__in=offering_ids).exclude(id__in=resource_with_usages)
    result = []

    for resource in resources_without_usages:
        if filter(lambda x: x['customer'] == resource.offering.customer, result):
            filter(lambda x: x['customer'] == resource.offering.customer, result)[0]['resources'].append(resource)
        else:
            result.append({
                'customer': resource.offering.customer,
                'resources': [resource],
            })

    return result


def get_public_resources_url(customer):
    link_template = settings.WALDUR_MARKETPLACE['PUBLIC_RESOURCES_LINK_TEMPLATE']
    return link_template.format(organization_uuid=customer.uuid)


def validate_limits(limits, offering):
    usage_components = offering.components \
        .filter(billing_type=models.OfferingComponent.BillingTypes.USAGE) \
        .exclude(disable_quotas=True) \
        .values_list('type', flat=True)
    valid_component_types = set(usage_components)
    valid_component_types.update(plugins.manager.get_available_limits(offering.type))
    invalid_types = set(limits.keys()) - valid_component_types
    if invalid_types:
        raise serializers.ValidationError({'limits': _('Invalid types: %s') % ', '.join(invalid_types)})

    # Validate max and min limit value.
    components_map = {
        component.type: component
        for component in offering.components.filter(type__in=valid_component_types)
    }

    for key, value in limits.items():
        component = components_map.get(key)
        if not component:
            continue

        if component.max_value and value > component.max_value:
            raise serializers.ValidationError(_('The limit %s value cannot be more than %s.') % (
                value, component.max_value
            ))
        if component.min_value and value < component.min_value:
            raise serializers.ValidationError(_('The limit %s value cannot be less than %s.') % (
                value, component.min_value
            ))


def create_offering_components(offering, custom_components=None):
    fixed_components = plugins.manager.get_components(offering.type)
    category_components = {
        component.type: component
        for component in models.CategoryComponent.objects.filter(category=offering.category)
    }

    for component_data in fixed_components:
        models.OfferingComponent.objects.create(
            offering=offering,
            parent=category_components.get(component_data.type, None),
            **component_data._asdict()
        )

    if custom_components:
        for component_data in custom_components:
            models.OfferingComponent.objects.create(offering=offering, **component_data)



from __future__ import unicode_literals

import functools
import importlib
import logging

from django.conf import settings
from django.utils.encoding import force_text
from django.utils.lru_cache import lru_cache
from rest_framework.reverse import reverse
import six

logger = logging.getLogger(__name__)


default_app_config = 'waldur_core.structure.apps.StructureConfig'


class SupportedServices(object):
    """ Comprehensive list of currently supported services and resources.
        Build the list via serializers definition on application start.
        Example data structure of registry:

        {
            'gitlab': {
                'name': 'GitLab',
                'model_name': 'gitlab.gitlabservice',
                'backend': nodeconductor_gitlab.backend.GitLabBackend,
                'detail_view': 'gitlab-detail',
                'list_view': 'gitlab-list',
                'properties': {},
                'resources': {
                    'gitlab.group': {
                        'name': 'Group',
                        'detail_view': 'gitlab-group-detail',
                        'list_view': 'gitlab-group-list'
                    },
                    'gitlab.project': {
                        'name': 'Project',
                        'detail_view': 'gitlab-project-detail',
                        'list_view': 'gitlab-project-list'
                    }
                }
            }
        }

    """

    @classmethod
    def get_filter_mapping(cls):
        return {name: code for code, name in cls.get_choices()}

    _registry = {}

    @classmethod
    def _setdefault(cls, service_key):
        cls._registry.setdefault(service_key, {
            'resources': {},
            'properties': {}
        })

    @classmethod
    def register_backend(cls, backend_class, nested=False):
        if not cls._is_active_model(backend_class):
            return

        # For nested backends just discover resources/properties
        if not nested:
            key = cls.get_model_key(backend_class)
            cls._setdefault(key)
            cls._registry[key]['backend'] = backend_class

        # Forcely import service serialize to run services autodiscovery
        try:
            module_name = backend_class.__module__
            importlib.import_module(module_name.replace('backend', 'serializers'))
        except ImportError:
            pass

    @classmethod
    def register_service(cls, model):
        if model is NotImplemented or not cls._is_active_model(model):
            return
        key = cls.get_model_key(model)
        cls._setdefault(key)
        cls._registry[key]['name'] = key
        cls._registry[key]['model_name'] = cls._get_model_str(model)
        cls._registry[key]['detail_view'] = cls.get_detail_view_for_model(model)
        cls._registry[key]['list_view'] = cls.get_list_view_for_model(model)

    @classmethod
    def register_service_serializer(cls, model, serializer):
        if model is NotImplemented or not cls._is_active_model(model):
            return
        key = cls.get_model_key(model)
        cls._setdefault(key)
        cls._registry[key]['serializer'] = serializer

    @classmethod
    def register_service_filter(cls, model, filter):
        if model is NotImplemented or not cls._is_active_model(model):
            return
        key = cls.get_model_key(model)
        cls._setdefault(key)
        cls._registry[key]['filter'] = filter

    @classmethod
    def register_resource_serializer(cls, model, serializer):
        if model is NotImplemented or not cls._is_active_model(model):
            return
        key = cls.get_model_key(model)
        cls._setdefault(key)
        model_str = cls._get_model_str(model)
        cls._registry[key]['resources'].setdefault(model_str, {'name': model.__name__})
        cls._registry[key]['resources'][model_str].setdefault('detail_view', cls.get_detail_view_for_model(model))
        cls._registry[key]['resources'][model_str].setdefault('list_view', cls.get_list_view_for_model(model))
        cls._registry[key]['resources'][model_str].setdefault('serializer', serializer)

    @classmethod
    def register_resource_filter(cls, model, filter):
        if model is NotImplemented or not cls._is_active_model(model) or model._meta.abstract:
            return
        key = cls.get_model_key(model)
        cls._setdefault(key)
        model_str = cls._get_model_str(model)
        cls._registry[key]['resources'].setdefault(model_str, {'name': model.__name__})
        cls._registry[key]['resources'][model_str]['filter'] = filter

    @classmethod
    def register_resource_view(cls, model, view):
        if model is NotImplemented or not cls._is_active_model(model) or model._meta.abstract:
            return
        key = cls.get_model_key(model)
        cls._setdefault(key)
        model_str = cls._get_model_str(model)
        cls._registry[key]['resources'].setdefault(model_str, {'name': model.__name__})
        cls._registry[key]['resources'][model_str]['view'] = view

    @classmethod
    def register_property(cls, model):
        if model is NotImplemented or not cls._is_active_model(model):
            return
        key = cls.get_model_key(model)
        cls._setdefault(key)
        model_str = cls._get_model_str(model)
        cls._registry[key]['properties'][model_str] = {
            'name': model.__name__,
            'list_view': cls.get_list_view_for_model(model)
        }

    @classmethod
    def get_service_backend(cls, key):
        if not isinstance(key, six.string_types):
            key = cls.get_model_key(key)
        try:
            return cls._registry[key]['backend']
        except KeyError:
            raise ServiceBackendNotImplemented

    @classmethod
    def get_services(cls, request=None):
        """ Get a list of services endpoints.
            {
                "Oracle": "/api/oracle/",
                "OpenStack": "/api/openstack/",
                "GitLab": "/api/gitlab/",
                "DigitalOcean": "/api/digitalocean/"
            }
        """
        return {service['name']: reverse(service['list_view'], request=request)
                for service in cls._registry.values()}

    @classmethod
    def get_service_serializer(cls, model):
        key = cls.get_model_key(model)
        return cls.get_service_serializer_for_key(key)

    @classmethod
    def get_service_serializer_for_key(cls, key):
        return cls._registry[key]['serializer']

    @classmethod
    def get_service_filter(cls, model):
        key = cls.get_model_key(model)
        return cls._registry[key]['filter']

    @classmethod
    def get_resources(cls, request=None):
        """ Get a list of resources endpoints.
            {
                "DigitalOcean.Droplet": "/api/digitalocean-droplets/",
                "Oracle.Database": "/api/oracle-databases/",
                "GitLab.Group": "/api/gitlab-groups/",
                "GitLab.Project": "/api/gitlab-projects/"
            }
        """
        return {'.'.join([service['name'], resource['name']]): reverse(resource['list_view'], request=request)
                for service in cls._registry.values()
                for resource in service['resources'].values()}

    @classmethod
    def get_resource_serializer(cls, model):
        key = cls.get_model_key(model)
        model_str = cls._get_model_str(model)
        return cls._registry[key]['resources'][model_str]['serializer']

    @classmethod
    def get_resource_filter(cls, model):
        key = cls.get_model_key(model)
        model_str = cls._get_model_str(model)
        return cls._registry[key]['resources'][model_str]['filter']

    @classmethod
    def get_resource_view(cls, model):
        key = cls.get_model_key(model)
        model_str = cls._get_model_str(model)
        return cls._registry[key]['resources'][model_str]['view']

    @classmethod
    def get_services_with_resources(cls, request=None):
        """ Get a list of services and resources endpoints.
            {
                ...
                "GitLab": {
                    "url": "/api/gitlab/",
                    "service_project_link_url": "/api/gitlab-service-project-link/",
                    "resources": {
                        "Project": "/api/gitlab-projects/",
                        "Group": "/api/gitlab-groups/"
                    }
                },
                ...
            }
        """
        from django.apps import apps

        data = {}
        for service in cls._registry.values():
            service_model = apps.get_model(service['model_name'])
            service_project_link = service_model.projects.through
            service_project_link_url = reverse(cls.get_list_view_for_model(service_project_link), request=request)

            data[service['name']] = {
                'url': reverse(service['list_view'], request=request),
                'service_project_link_url': service_project_link_url,
                'resources': {resource['name']: reverse(resource['list_view'], request=request)
                              for resource in service['resources'].values()},
                'properties': {resource['name']: reverse(resource['list_view'], request=request)
                               for resource in service.get('properties', {}).values()},
                'is_public_service': cls.is_public_service(service_model)
            }
        return data

    @classmethod
    @lru_cache(maxsize=1)
    def get_service_models(cls):
        """ Get a list of service models.
            {
                ...
                'gitlab': {
                    "service": nodeconductor_gitlab.models.GitLabService,
                    "service_project_link": nodeconductor_gitlab.models.GitLabServiceProjectLink,
                    "resources": [
                        nodeconductor_gitlab.models.Group,
                        nodeconductor_gitlab.models.Project
                    ],
                },
                ...
            }

        """
        from django.apps import apps

        data = {}
        for key, service in cls._registry.items():
            service_model = apps.get_model(service['model_name'])
            service_project_link = service_model.projects.through
            data[key] = {
                'service': service_model,
                'service_project_link': service_project_link,
                'resources': [apps.get_model(r) for r in service['resources'].keys()],
                'properties': [apps.get_model(r) for r in service['properties'].keys() if '.' in r],
            }

        return data

    @classmethod
    @lru_cache(maxsize=1)
    def get_resource_models(cls):
        """ Get a list of resource models.
            {
                'DigitalOcean.Droplet': waldur_digitalocean.models.Droplet,
                'JIRA.Project': waldur_jira.models.Project,
                'OpenStack.Tenant': waldur_openstack.models.Tenant
            }

        """
        from django.apps import apps

        return {'.'.join([service['name'], attrs['name']]): apps.get_model(resource)
                for service in cls._registry.values()
                for resource, attrs in service['resources'].items()}

    @classmethod
    @lru_cache(maxsize=20)
    def get_service_resources(cls, model):
        """ Get resource models by service model """
        key = cls.get_model_key(model)
        return cls.get_service_name_resources(key)

    @classmethod
    @lru_cache(maxsize=20)
    def get_resource_serializers(cls):
        return [resource['serializer']
                for provider in cls._registry.values()
                for resource in provider['resources'].values()]

    @classmethod
    @lru_cache(maxsize=20)
    def get_service_name_resources(cls, service_name):
        """ Get resource models by service name """
        from django.apps import apps

        resources = cls._registry[service_name]['resources'].keys()
        return [apps.get_model(resource) for resource in resources]

    @classmethod
    def get_name_for_model(cls, model):
        """ Get a name for given class or model:
            -- it's a service type for a service
            -- it's a <service_type>.<resource_model_name> for a resource
        """
        key = cls.get_model_key(model)
        model_str = cls._get_model_str(model)
        service = cls._registry[key]
        if model_str in service['resources']:
            return '{}.{}'.format(service['name'], service['resources'][model_str]['name'])
        else:
            return service['name']

    @classmethod
    def get_related_models(cls, model):
        """ Get a dictionary with related structure models for given class or model:

            >> SupportedServices.get_related_models(gitlab_models.Project)
            {
                'service': nodeconductor_gitlab.models.GitLabService,
                'service_project_link': nodeconductor_gitlab.models.GitLabServiceProjectLink,
                'resources': [
                    nodeconductor_gitlab.models.Group,
                    nodeconductor_gitlab.models.Project,
                ]
            }
        """
        from waldur_core.structure.models import ServiceSettings

        if isinstance(model, ServiceSettings):
            model_str = cls._registry.get(model.type, {}).get('model_name', '')
        else:
            model_str = cls._get_model_str(model)

        for models in cls.get_service_models().values():
            if model_str == cls._get_model_str(models['service']) or \
               model_str == cls._get_model_str(models['service_project_link']):
                return models

            for resource_model in models['resources']:
                if model_str == cls._get_model_str(resource_model):
                    return models

    @classmethod
    def _is_active_model(cls, model):
        """ Check is model app name is in list of INSTALLED_APPS """
        # We need to use such tricky way to check because of inconsistent apps names:
        # some apps are included in format "<module_name>.<app_name>" like "waldur_core.openstack"
        # other apps are included in format "<app_name>" like "nodecondcutor_sugarcrm"
        return ('.'.join(model.__module__.split('.')[:2]) in settings.INSTALLED_APPS or
                '.'.join(model.__module__.split('.')[:1]) in settings.INSTALLED_APPS)

    @classmethod
    def _get_model_str(cls, model):
        return force_text(model._meta)

    @classmethod
    def get_model_key(cls, model):
        return cls.get_app_config(model).service_name

    @classmethod
    def is_public_service(cls, model):
        return getattr(cls.get_app_config(model), 'is_public_service', False)

    @classmethod
    def get_app_config(cls, model):
        from django.apps import apps
        return apps.get_containing_app_config(model.__module__)

    @classmethod
    def get_list_view_for_model(cls, model):
        return model.get_url_name() + '-list'

    @classmethod
    def get_detail_view_for_model(cls, model):
        return model.get_url_name() + '-detail'

    @classmethod
    @lru_cache(maxsize=1)
    def get_choices(cls):
        items = [(code, service['name']) for code, service in cls._registry.items()]
        return sorted(items, key=lambda pair: pair[1])

    @classmethod
    def has_service_type(cls, service_type):
        return service_type in cls._registry

    @classmethod
    def get_name_for_type(cls, service_type):
        try:
            return cls._registry[service_type]['name']
        except KeyError:
            return service_type


class ServiceBackendError(Exception):
    """ Base exception for errors occurring during backend communication. """
    pass


def log_backend_action(action=None):
    """ Logging for backend method.

    Expects django model instance as first argument.
    """
    def decorator(func):
        @functools.wraps(func)
        def wrapped(self, instance, *args, **kwargs):
            action_name = func.func_name.replace('_', ' ') if action is None else action

            logger.debug('About to %s `%s` (PK: %s).', action_name, instance, instance.pk)
            result = func(self, instance, *args, **kwargs)
            logger.debug('Action `%s` was executed successfully for `%s` (PK: %s).',
                         action_name, instance, instance.pk)
            return result
        return wrapped
    return decorator


class ServiceBackendNotImplemented(NotImplementedError):
    pass


class ServiceBackend(object):
    """ Basic service backed with only common methods pre-defined. """

    DEFAULTS = {}

    def __init__(self, settings, **kwargs):
        pass

    def ping(self, raise_exception=False):
        raise ServiceBackendNotImplemented

    def sync(self):
        self.pull_service_properties()
        self.pull_resources()
        self.pull_subresources()

    def pull_service_properties(self):
        pass

    def pull_resources(self):
        pass

    def pull_subresources(self):
        pass

    def has_global_properties(self):
        return False

    def get_resources_for_import(self):
        raise ServiceBackendNotImplemented

    def get_managed_resources(self):
        raise ServiceBackendNotImplemented

    def get_monthly_cost_estimate(self, resource):
        raise ServiceBackendNotImplemented

    def get_stats(self):
        raise ServiceBackendNotImplemented

    @staticmethod
    def gb2mb(val):
        return int(val * 1024) if val else 0

    @staticmethod
    def tb2mb(val):
        return int(val * 1024 * 1024) if val else 0

    @staticmethod
    def mb2gb(val):
        return val / 1024 if val else 0

    @staticmethod
    def mb2tb(val):
        return val / 1024 / 1024 if val else 0



from __future__ import unicode_literals

from django.db import models
from django.conf import settings


class AuthProfile(models.Model):
    user = models.OneToOneField(settings.AUTH_USER_MODEL, related_name='auth_profile')
    google = models.CharField(max_length=120, blank=True, null=True, unique=True)
    facebook = models.CharField(max_length=120, blank=True, null=True, unique=True)



from waldur_core.core import WaldurExtension


class SlurmExtension(WaldurExtension):
    class Settings:
        WALDUR_SLURM = {
            'ENABLED': False,
            'CUSTOMER_PREFIX': 'waldur_customer_',
            'PROJECT_PREFIX': 'waldur_project_',
            'ALLOCATION_PREFIX': 'waldur_allocation_',
            'PRIVATE_KEY_PATH': '/etc/waldur/id_rsa',
        }

    @staticmethod
    def django_app():
        return 'waldur_slurm'

    @staticmethod
    def rest_urls():
        from .urls import register_in
        return register_in

    @staticmethod
    def get_cleanup_executor():
        from waldur_slurm.executors import SlurmCleanupExecutor
        return SlurmCleanupExecutor



# -*- coding: utf-8 -*-
# Generated by Django 1.11.7 on 2018-01-16 08:08
from __future__ import unicode_literals

from django.db import migrations, models
import django.db.models.deletion
import django.utils.timezone
import model_utils.fields


class Migration(migrations.Migration):

    dependencies = [
        ('structure', '0053_add_project_type'),
        ('openstack', '0036_remove_security_group_duplicates'),
    ]

    operations = [
        migrations.CreateModel(
            name='CustomerOpenStack',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('created', model_utils.fields.AutoCreatedField(default=django.utils.timezone.now, editable=False, verbose_name='created')),
                ('modified', model_utils.fields.AutoLastModifiedField(default=django.utils.timezone.now, editable=False, verbose_name='modified')),
                ('external_network_id', models.CharField(max_length=255, verbose_name='OpenStack external network ID')),
                ('customer', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='structure.Customer')),
                ('settings', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='structure.ServiceSettings')),
            ],
            options={
                'verbose_name': 'Organization OpenStack settings',
                'verbose_name_plural': 'Organization OpenStack settings',
            },
        ),
        migrations.AlterUniqueTogether(
            name='customeropenstack',
            unique_together=set([('settings', 'customer')]),
        ),
    ]



# -*- coding: utf-8 -*-
# Generated by Django 1.11.7 on 2018-01-22 13:07
from __future__ import unicode_literals

import logging

from django.db import migrations

from waldur_core.core import fields


logger = logging.getLogger(__name__)


def migrate_data(apps, schema_editor):
    PaymentDetails = apps.get_model('invoices', 'PaymentDetails')
    properties = [
        ['_company', 'name'],
        ['_type', 'type'],
        ['_address', 'address'],
        ['_country', 'country'],
        ['_email', 'email'],
        ['_postal', 'postal'],
        ['_phone', 'phone_number'],
        ['_bank', 'bank_name'],
        ['_account', 'bank_account'],
        ['_accounting_start_date', 'accounting_start_date'],
        ['_default_tax_percent', 'default_tax_percent'],
    ]
    
    for pd in PaymentDetails.objects.all():
        
        for p in properties:
            value_payment = getattr(pd, p[0])
            value_customer = getattr(pd.customer, p[1])

            if p[1] in ['accounting_start_date', 'default_tax_percent'] and value_payment:
                setattr(pd.customer, p[1], value_payment)
                continue

            if not value_customer and value_payment:
                if not p[1] == 'country':
                    setattr(pd.customer, p[1], value_payment)
                else:
                    country = filter(lambda x: x[1] == value_payment, fields.CountryField.COUNTRIES)
                    if country:
                        setattr(pd.customer, p[1], country[0][0])
                    else:
                        logger.warning('Migrate payment details data. Invalid country name %s' % value_payment)
                
        pd.customer.save()


class Migration(migrations.Migration):

    dependencies = [
        ('invoices', '0019_mark_payments_details_fields'),
        ('structure', '0054_payment_details'),
    ]

    operations = [
        migrations.RunPython(migrate_data, reverse_code=migrations.RunPython.noop, elidable=True),
    ]



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from django.db import models, migrations


class Migration(migrations.Migration):

    dependencies = [
        ('structure', '0032_make_options_optional'),
    ]

    operations = [
        migrations.RemoveField(
            model_name='servicesettings',
            name='dummy',
        ),
    ]



from django.conf.urls import url

from waldur_core.media import views


urlpatterns = [
    url(r'^media-download/(?P<token>.+)/$',
        views.ProtectedFileView.as_view(),
        name='media-download'),
]



import logging

from django.core.exceptions import ObjectDoesNotExist
from django.db import transaction
from django.template import Context, Template

from waldur_mastermind.marketplace import callbacks
from waldur_mastermind.marketplace import models as marketplace_models
from waldur_mastermind.marketplace_support import PLUGIN_NAME
from waldur_mastermind.support import models as support_models

from . import tasks

logger = logging.getLogger(__name__)


def create_support_template(sender, instance, created=False, **kwargs):
    if instance.type != PLUGIN_NAME or not created:
        return

    if not instance.scope:
        template = support_models.OfferingTemplate.objects.create(
            name=instance.name,
            config=instance.options
        )
        instance.scope = template
        instance.save()


def update_support_template(sender, instance, created=False, **kwargs):
    offering = instance

    if offering.type != PLUGIN_NAME:
        return

    if created:
        return

    if not offering.tracker.has_changed('options'):
        return

    if not offering.scope:
        return

    template = offering.scope
    template.config = offering.options
    template.save(update_fields=['config'])


def change_order_item_state(sender, instance, created=False, **kwargs):
    if created:
        return

    if not instance.tracker.has_changed('state'):
        return

    try:
        resource = marketplace_models.Resource.objects.get(scope=instance)
    except ObjectDoesNotExist:
        logger.warning('Skipping support offering state synchronization '
                       'because related order item is not found. Offering ID: %s', instance.id)
        return

    if instance.state == support_models.Offering.States.OK:
        callbacks.resource_creation_succeeded(resource)
    elif instance.state == support_models.Offering.States.TERMINATED:
        if instance.tracker.previous('state') == support_models.Offering.States.REQUESTED:
            callbacks.resource_creation_failed(resource)
        if instance.tracker.previous('state') == support_models.Offering.States.OK:
            callbacks.resource_deletion_succeeded(resource)


def terminate_resource(sender, instance, **kwargs):
    try:
        resource = marketplace_models.Resource.objects.get(scope=instance)
    except ObjectDoesNotExist:
        logger.debug('Skipping resource terminate for support request '
                     'because related resource does not exist. '
                     'Request ID: %s', instance.id)
    else:
        callbacks.resource_deletion_succeeded(resource)


def create_or_update_support_plan(sender, instance, created=False, **kwargs):
    plan = instance
    if plan.offering.type != PLUGIN_NAME:
        return

    if not created and not plan.tracker.has_changed('unit_price'):
        return

    if not isinstance(plan.offering.scope, support_models.OfferingTemplate):
        return

    with transaction.atomic():
        if not plan.scope:
            offering_plan = support_models.OfferingPlan.objects.create(
                template=plan.offering.scope,
                name=plan.name,
                description=plan.description,
                product_code=plan.product_code,
                article_code=plan.article_code,
                unit=plan.unit,
                unit_price=plan.unit_price,
            )
            plan.scope = offering_plan
            plan.save()
        else:
            plan.scope.unit_price = plan.unit_price
            plan.scope.save(update_fields=['unit_price'])


def change_offering_state(sender, instance, created=False, **kwargs):
    """ Processing of creating support offering issue."""
    if created:
        return

    issue = instance

    if instance.tracker.has_changed('status') and issue.resolved is not None:
        try:
            offering = support_models.Offering.objects.get(issue=issue)
        except support_models.Offering.DoesNotExist:
            logger.warning('Skipping issue state synchronization '
                           'because related support offering is not found. Issue ID: %s', issue.id)
            return

        if issue.resolved:
            offering.state = support_models.Offering.States.OK
        else:
            offering.state = support_models.Offering.States.TERMINATED

        offering.save()


def update_order_item_if_issue_was_complete(sender, instance, created=False, **kwargs):
    """ Processing of terminating or updating a resource."""
    if created:
        return

    issue = instance

    if not issue.tracker.has_changed('status'):
        return

    if issue.resource \
            and isinstance(issue.resource, marketplace_models.OrderItem) \
            and issue.resource.offering.type == PLUGIN_NAME and issue.resolved is not None:
        order_item = issue.resource

        if issue.resolved:
            request = order_item.resource.scope
            # A request is object of support.Offering.
            # Support.Offering object created from a marketplace is called 'request' in a frontend

            if not request:
                logger.warning('Skipping resource termination '
                               'because request is not found. Order item ID: %s', order_item.id)
                return

            with transaction.atomic():
                if order_item.type == marketplace_models.OrderItem.Types.TERMINATE:
                    request.delete()
                    # callbacks.resource_deletion_succeeded will called in terminate_resource handler
                elif order_item.type == marketplace_models.OrderItem.Types.UPDATE:
                    if request.issue != issue:
                        request.issue = issue
                        request.save(update_fields=['issue'])
                    callbacks.resource_update_succeeded(order_item.resource)
        else:
            with transaction.atomic():
                if order_item.type == marketplace_models.OrderItem.Types.TERMINATE:
                    callbacks.resource_deletion_failed(order_item.resource)
                elif order_item.type == marketplace_models.OrderItem.Types.UPDATE:
                    callbacks.resource_update_failed(order_item.resource)


def notify_about_request_based_item_creation(sender, instance, created=False, **kwargs):
    if created:
        return

    issue = instance

    if not issue.tracker.has_changed('backend_id'):
        return

    try:
        support_offering = support_models.Offering.objects.get(issue=issue)
    except support_models.Offering.DoesNotExist:
        return

    try:
        resource = marketplace_models.Resource.objects.get(scope=support_offering)
    except marketplace_models.Resource.DoesNotExist:
        return

    try:
        order_item = marketplace_models.OrderItem.objects.get(resource=resource,
                                                              type=marketplace_models.OrderItem.Types.CREATE)
    except marketplace_models.OrderItem.DoesNotExist:
        return

    if order_item.offering.type != PLUGIN_NAME:
        return

    service_provider = getattr(order_item.offering.customer, 'serviceprovider', None)

    if not service_provider:
        logger.warning('Customer providing an Offering is not registered as a Service Provider.')
        return

    if not service_provider.lead_email:
        return

    context = Context({'order_item': order_item, 'issue': issue}, autoescape=False)
    template = Template(service_provider.lead_body)
    message = template.render(context).strip()
    template = Template(service_provider.lead_subject)
    subject = template.render(context).strip()

    transaction.on_commit(lambda: tasks.send_mail_notification.delay(subject, message, service_provider.lead_email))



# -*- coding: utf-8 -*-
# Generated by Django 1.11.24 on 2019-09-10 10:28
from __future__ import unicode_literals

import django.core.validators
from django.db import migrations, models
import django.db.models.deletion
import django.utils.timezone
import django_fsm
import model_utils.fields
import taggit.managers
import waldur_core.core.fields
import waldur_core.core.models
import waldur_core.core.validators
import waldur_core.logging.loggers
import waldur_core.structure.models
import waldur_openstack.openstack_tenant.models


class Migration(migrations.Migration):

    replaces = [(b'openstack_tenant', '0001_initial'), (b'openstack_tenant', '0002_service_properties'), (b'openstack_tenant', '0003_volume_and_snapshot'), (b'openstack_tenant', '0004_instance'), (b'openstack_tenant', '0005_resources_actions'), (b'openstack_tenant', '0006_resource_action_details'), (b'openstack_tenant', '0007_backup_backuprestoration'), (b'openstack_tenant', '0008_backup_schedule'), (b'openstack_tenant', '0009_tenant_service_verbose_name'), (b'openstack_tenant', '0010_rename_floating_ip_status'), (b'openstack_tenant', '0011_backupschedule_call_count'), (b'openstack_tenant', '0012_backupschedule_resource'), (b'openstack_tenant', '0013_init_backupschedule_spl'), (b'openstack_tenant', '0014_make_backupschedule_spl_non_nullable'), (b'openstack_tenant', '0015_snapshotrestoration'), (b'openstack_tenant', '0016_network_subnet_internalip'), (b'openstack_tenant', '0017_snapshot_schedule'), (b'openstack_tenant', '0018_remove_openstacktenantservice_name'), (b'openstack_tenant', '0019_migrate_to_single_external_ip'), (b'openstack_tenant', '0020_remove_external_ips'), (b'openstack_tenant', '0021_instance_subnets'), (b'openstack_tenant', '0022_floatingip_internal_ip'), (b'openstack_tenant', '0023_remove_instance_external_ip'), (b'openstack_tenant', '0024_add_backup_size'), (b'openstack_tenant', '0025_copy_certifications_from_existing_settings'), (b'openstack_tenant', '0026_remove_start_time'), (b'openstack_tenant', '0027_remove_duplicate_floating_ips'), (b'openstack_tenant', '0028_remove_duplicate_security_groups_networks_subnets'), (b'openstack_tenant', '0029_add_unique_constraint_for_properties'), (b'openstack_tenant', '0030_add_volume_image_name'), (b'openstack_tenant', '0031_unique_internal_ip'), (b'openstack_tenant', '0032_nullable_internal_ip_instance'), (b'openstack_tenant', '0033_unique_instance_backend_id'), (b'openstack_tenant', '0034_immutable_default_json'), (b'openstack_tenant', '0035_unique_floating_ip'), (b'openstack_tenant', '0036_instance_shared_ips'), (b'openstack_tenant', '0037_internal_ip_settings'), (b'openstack_tenant', '0038_internal_ip_settings_non_null'), (b'openstack_tenant', '0039_forbid_blank_backend_id'), (b'openstack_tenant', '0040_volume_type'), (b'openstack_tenant', '0041_volume_type_field'), (b'openstack_tenant', '0042_volume_availability_zone'), (b'openstack_tenant', '0043_field_availability_zone'), (b'openstack_tenant', '0044_instance_availability_zone'), (b'openstack_tenant', '0045_volumeavailabilityzone_available'), (b'openstack_tenant', '0046_instanceavailabilityzone_available')]

    initial = True

    dependencies = [
        ('structure', '0009_project_is_removed'),
        ('taggit', '0002_auto_20150616_2121'),
    ]

    operations = [
        migrations.CreateModel(
            name='Backup',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('created', model_utils.fields.AutoCreatedField(default=django.utils.timezone.now, editable=False, verbose_name='created')),
                ('modified', model_utils.fields.AutoLastModifiedField(default=django.utils.timezone.now, editable=False, verbose_name='modified')),
                ('description', models.CharField(blank=True, max_length=500, verbose_name='description')),
                ('name', models.CharField(max_length=150, validators=[waldur_core.core.validators.validate_name], verbose_name='name')),
                ('uuid', waldur_core.core.fields.UUIDField()),
                ('error_message', models.TextField(blank=True)),
                ('state', django_fsm.FSMIntegerField(choices=[(5, 'Creation Scheduled'), (6, 'Creating'), (1, 'Update Scheduled'), (2, 'Updating'), (7, 'Deletion Scheduled'), (8, 'Deleting'), (3, 'OK'), (4, 'Erred')], default=5)),
                ('backend_id', models.CharField(blank=True, max_length=255)),
                ('kept_until', models.DateTimeField(blank=True, help_text='Guaranteed time of backup retention. If null - keep forever.', null=True)),
                ('metadata', waldur_core.core.fields.JSONField(blank=True, help_text='Additional information about backup, can be used for backup restoration or deletion')),
            ],
            options={
                'abstract': False,
            },
            bases=(waldur_core.core.models.DescendantMixin, waldur_core.core.models.BackendModelMixin, waldur_core.structure.models.StructureLoggableMixin, models.Model),
        ),
        migrations.CreateModel(
            name='BackupRestoration',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('created', model_utils.fields.AutoCreatedField(default=django.utils.timezone.now, editable=False, verbose_name='created')),
                ('modified', model_utils.fields.AutoLastModifiedField(default=django.utils.timezone.now, editable=False, verbose_name='modified')),
                ('uuid', waldur_core.core.fields.UUIDField()),
                ('backup', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='restorations', to='openstack_tenant.Backup')),
            ],
            options={
                'abstract': False,
            },
        ),
        migrations.CreateModel(
            name='BackupSchedule',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('created', model_utils.fields.AutoCreatedField(default=django.utils.timezone.now, editable=False, verbose_name='created')),
                ('modified', model_utils.fields.AutoLastModifiedField(default=django.utils.timezone.now, editable=False, verbose_name='modified')),
                ('description', models.CharField(blank=True, max_length=500, verbose_name='description')),
                ('name', models.CharField(max_length=150, validators=[waldur_core.core.validators.validate_name], verbose_name='name')),
                ('uuid', waldur_core.core.fields.UUIDField()),
                ('error_message', models.TextField(blank=True)),
                ('schedule', waldur_core.core.fields.CronScheduleField(max_length=15, validators=[waldur_core.core.validators.validate_cron_schedule, waldur_core.core.validators.MinCronValueValidator(1)])),
                ('next_trigger_at', models.DateTimeField(null=True)),
                ('timezone', models.CharField(default=django.utils.timezone.get_current_timezone_name, max_length=50)),
                ('is_active', models.BooleanField(default=False)),
                ('state', django_fsm.FSMIntegerField(choices=[(5, 'Creation Scheduled'), (6, 'Creating'), (1, 'Update Scheduled'), (2, 'Updating'), (7, 'Deletion Scheduled'), (8, 'Deleting'), (3, 'OK'), (4, 'Erred')], default=5)),
                ('backend_id', models.CharField(blank=True, max_length=255)),
                ('retention_time', models.PositiveIntegerField(help_text='Retention time in days, if 0 - resource will be kept forever')),
                ('maximal_number_of_resources', models.PositiveSmallIntegerField()),
                ('call_count', models.PositiveSmallIntegerField(default=0, help_text='How many times a resource schedule was called.')),
            ],
            options={
                'abstract': False,
            },
            bases=(waldur_core.core.models.DescendantMixin, waldur_core.core.models.BackendModelMixin, waldur_core.structure.models.StructureLoggableMixin, models.Model),
        ),
        migrations.CreateModel(
            name='Flavor',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('name', models.CharField(max_length=150, validators=[waldur_core.core.validators.validate_name], verbose_name='name')),
                ('uuid', waldur_core.core.fields.UUIDField()),
                ('backend_id', models.CharField(db_index=True, max_length=255)),
                ('cores', models.PositiveSmallIntegerField(help_text='Number of cores in a VM')),
                ('ram', models.PositiveIntegerField(help_text='Memory size in MiB')),
                ('disk', models.PositiveIntegerField(help_text='Root disk size in MiB')),
                ('settings', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='+', to='structure.ServiceSettings')),
            ],
            bases=(waldur_core.logging.loggers.LoggableMixin, waldur_core.core.models.BackendModelMixin, models.Model),
        ),
        migrations.CreateModel(
            name='FloatingIP',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('name', models.CharField(max_length=150, validators=[waldur_core.core.validators.validate_name], verbose_name='name')),
                ('uuid', waldur_core.core.fields.UUIDField()),
                ('backend_id', models.CharField(db_index=True, max_length=255)),
                ('address', models.GenericIPAddressField(default=None, null=True, protocol='IPv4')),
                ('runtime_state', models.CharField(max_length=30)),
                ('backend_network_id', models.CharField(editable=False, max_length=255)),
                ('is_booked', models.BooleanField(default=False, help_text='Marks if floating IP has been booked for provisioning.')),
            ],
            options={
                'verbose_name': 'Floating IP',
                'verbose_name_plural': 'Floating IPs',
            },
            bases=(waldur_core.core.models.BackendModelMixin, models.Model),
        ),
        migrations.CreateModel(
            name='Image',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('name', models.CharField(max_length=150, validators=[waldur_core.core.validators.validate_name], verbose_name='name')),
                ('uuid', waldur_core.core.fields.UUIDField()),
                ('backend_id', models.CharField(db_index=True, max_length=255)),
                ('min_disk', models.PositiveIntegerField(default=0, help_text='Minimum disk size in MiB')),
                ('min_ram', models.PositiveIntegerField(default=0, help_text='Minimum memory size in MiB')),
                ('settings', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='+', to='structure.ServiceSettings')),
            ],
            options={
                'abstract': False,
            },
            bases=(waldur_core.core.models.BackendModelMixin, models.Model),
        ),
        migrations.CreateModel(
            name='Instance',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('created', model_utils.fields.AutoCreatedField(default=django.utils.timezone.now, editable=False, verbose_name='created')),
                ('modified', model_utils.fields.AutoLastModifiedField(default=django.utils.timezone.now, editable=False, verbose_name='modified')),
                ('description', models.CharField(blank=True, max_length=500, verbose_name='description')),
                ('name', models.CharField(max_length=150, validators=[waldur_core.core.validators.validate_name], verbose_name='name')),
                ('uuid', waldur_core.core.fields.UUIDField()),
                ('error_message', models.TextField(blank=True)),
                ('latitude', models.FloatField(blank=True, null=True)),
                ('longitude', models.FloatField(blank=True, null=True)),
                ('runtime_state', models.CharField(blank=True, max_length=150, verbose_name='runtime state')),
                ('state', django_fsm.FSMIntegerField(choices=[(5, 'Creation Scheduled'), (6, 'Creating'), (1, 'Update Scheduled'), (2, 'Updating'), (7, 'Deletion Scheduled'), (8, 'Deleting'), (3, 'OK'), (4, 'Erred')], default=5)),
                ('cores', models.PositiveSmallIntegerField(default=0, help_text='Number of cores in a VM')),
                ('ram', models.PositiveIntegerField(default=0, help_text='Memory size in MiB')),
                ('disk', models.PositiveIntegerField(default=0, help_text='Disk size in MiB')),
                ('min_ram', models.PositiveIntegerField(default=0, help_text='Minimum memory size in MiB')),
                ('min_disk', models.PositiveIntegerField(default=0, help_text='Minimum disk size in MiB')),
                ('image_name', models.CharField(blank=True, max_length=150)),
                ('key_name', models.CharField(blank=True, max_length=50)),
                ('key_fingerprint', models.CharField(blank=True, max_length=47)),
                ('user_data', models.TextField(blank=True, help_text='Additional data that will be added to instance on provisioning')),
                ('start_time', models.DateTimeField(blank=True, null=True)),
                ('backend_id', models.CharField(blank=True, max_length=255, null=True)),
                ('flavor_name', models.CharField(blank=True, max_length=255)),
                ('flavor_disk', models.PositiveIntegerField(default=0, help_text='Flavor disk size in MiB')),
                ('action', models.CharField(blank=True, max_length=50)),
                ('action_details', waldur_core.core.fields.JSONField(default=dict)),
            ],
            bases=(waldur_openstack.openstack_tenant.models.TenantQuotaMixin, waldur_core.core.models.DescendantMixin, waldur_core.core.models.BackendModelMixin, waldur_core.structure.models.StructureLoggableMixin, models.Model),
        ),
        migrations.CreateModel(
            name='InstanceAvailabilityZone',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('name', models.CharField(max_length=150, validators=[waldur_core.core.validators.validate_name], verbose_name='name')),
                ('uuid', waldur_core.core.fields.UUIDField()),
                ('available', models.BooleanField(default=True)),
                ('settings', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='+', to='structure.ServiceSettings')),
            ],
            bases=(waldur_core.core.models.BackendModelMixin, models.Model),
        ),
        migrations.CreateModel(
            name='InternalIP',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('mac_address', models.CharField(blank=True, max_length=32)),
                ('ip4_address', models.GenericIPAddressField(blank=True, null=True, protocol=b'IPv4')),
                ('ip6_address', models.GenericIPAddressField(blank=True, null=True, protocol=b'IPv6')),
                ('backend_id', models.CharField(max_length=255, null=True)),
                ('instance', models.ForeignKey(null=True, on_delete=django.db.models.deletion.CASCADE, related_name='internal_ips_set', to='openstack_tenant.Instance')),
                ('settings', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='+', to='structure.ServiceSettings')),
            ],
            bases=(waldur_core.core.models.BackendModelMixin, models.Model),
        ),
        migrations.CreateModel(
            name='Network',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('description', models.CharField(blank=True, max_length=500, verbose_name='description')),
                ('name', models.CharField(max_length=150, validators=[waldur_core.core.validators.validate_name], verbose_name='name')),
                ('uuid', waldur_core.core.fields.UUIDField()),
                ('backend_id', models.CharField(db_index=True, max_length=255)),
                ('is_external', models.BooleanField(default=False)),
                ('type', models.CharField(blank=True, max_length=50)),
                ('segmentation_id', models.IntegerField(null=True)),
                ('settings', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='+', to='structure.ServiceSettings')),
            ],
            bases=(waldur_core.core.models.BackendModelMixin, models.Model),
        ),
        migrations.CreateModel(
            name='OpenStackTenantService',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('uuid', waldur_core.core.fields.UUIDField()),
                ('available_for_all', models.BooleanField(default=False, help_text='Service will be automatically added to all customers projects if it is available for all')),
                ('customer', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='structure.Customer', verbose_name='organization')),
            ],
            options={
                'verbose_name': 'OpenStackTenant provider',
                'verbose_name_plural': 'OpenStackTenant providers',
            },
            bases=(waldur_core.core.models.DescendantMixin, waldur_core.structure.models.StructureLoggableMixin, models.Model),
        ),
        migrations.CreateModel(
            name='OpenStackTenantServiceProjectLink',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('project', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='structure.Project')),
                ('service', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='openstack_tenant.OpenStackTenantService')),
            ],
            options={
                'abstract': False,
                'verbose_name': 'OpenStackTenant provider project link',
                'verbose_name_plural': 'OpenStackTenant provider project links',
            },
            bases=(waldur_core.core.models.DescendantMixin, waldur_core.logging.loggers.LoggableMixin, models.Model),
        ),
        migrations.CreateModel(
            name='SecurityGroup',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('description', models.CharField(blank=True, max_length=500, verbose_name='description')),
                ('name', models.CharField(max_length=150, validators=[waldur_core.core.validators.validate_name], verbose_name='name')),
                ('uuid', waldur_core.core.fields.UUIDField()),
                ('backend_id', models.CharField(db_index=True, max_length=255)),
                ('settings', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='+', to='structure.ServiceSettings')),
            ],
            bases=(waldur_core.core.models.BackendModelMixin, models.Model),
        ),
        migrations.CreateModel(
            name='SecurityGroupRule',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('protocol', models.CharField(blank=True, choices=[(b'tcp', b'tcp'), (b'udp', b'udp'), (b'icmp', b'icmp')], max_length=4)),
                ('from_port', models.IntegerField(null=True, validators=[django.core.validators.MaxValueValidator(65535)])),
                ('to_port', models.IntegerField(null=True, validators=[django.core.validators.MaxValueValidator(65535)])),
                ('cidr', models.CharField(blank=True, max_length=32)),
                ('backend_id', models.CharField(blank=True, max_length=128)),
                ('security_group', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='rules', to='openstack_tenant.SecurityGroup')),
            ],
            options={
                'abstract': False,
            },
        ),
        migrations.CreateModel(
            name='Snapshot',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('created', model_utils.fields.AutoCreatedField(default=django.utils.timezone.now, editable=False, verbose_name='created')),
                ('modified', model_utils.fields.AutoLastModifiedField(default=django.utils.timezone.now, editable=False, verbose_name='modified')),
                ('description', models.CharField(blank=True, max_length=500, verbose_name='description')),
                ('name', models.CharField(max_length=150, validators=[waldur_core.core.validators.validate_name], verbose_name='name')),
                ('uuid', waldur_core.core.fields.UUIDField()),
                ('error_message', models.TextField(blank=True)),
                ('runtime_state', models.CharField(blank=True, max_length=150, verbose_name='runtime state')),
                ('state', django_fsm.FSMIntegerField(choices=[(5, 'Creation Scheduled'), (6, 'Creating'), (1, 'Update Scheduled'), (2, 'Updating'), (7, 'Deletion Scheduled'), (8, 'Deleting'), (3, 'OK'), (4, 'Erred')], default=5)),
                ('size', models.PositiveIntegerField(help_text='Size in MiB')),
                ('backend_id', models.CharField(blank=True, max_length=255, null=True)),
                ('metadata', waldur_core.core.fields.JSONField(blank=True)),
                ('action', models.CharField(blank=True, max_length=50)),
                ('action_details', waldur_core.core.fields.JSONField(default=dict)),
                ('kept_until', models.DateTimeField(blank=True, help_text='Guaranteed time of snapshot retention. If null - keep forever.', null=True)),
                ('service_project_link', models.ForeignKey(on_delete=django.db.models.deletion.PROTECT, related_name='snapshots', to='openstack_tenant.OpenStackTenantServiceProjectLink')),
            ],
            bases=(waldur_openstack.openstack_tenant.models.TenantQuotaMixin, waldur_core.core.models.DescendantMixin, waldur_core.core.models.BackendModelMixin, waldur_core.structure.models.StructureLoggableMixin, models.Model),
        ),
        migrations.CreateModel(
            name='SnapshotRestoration',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('created', model_utils.fields.AutoCreatedField(default=django.utils.timezone.now, editable=False, verbose_name='created')),
                ('modified', model_utils.fields.AutoLastModifiedField(default=django.utils.timezone.now, editable=False, verbose_name='modified')),
                ('uuid', waldur_core.core.fields.UUIDField()),
                ('snapshot', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='restorations', to='openstack_tenant.Snapshot')),
            ],
            options={
                'abstract': False,
            },
        ),
        migrations.CreateModel(
            name='SnapshotSchedule',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('created', model_utils.fields.AutoCreatedField(default=django.utils.timezone.now, editable=False, verbose_name='created')),
                ('modified', model_utils.fields.AutoLastModifiedField(default=django.utils.timezone.now, editable=False, verbose_name='modified')),
                ('description', models.CharField(blank=True, max_length=500, verbose_name='description')),
                ('name', models.CharField(max_length=150, validators=[waldur_core.core.validators.validate_name], verbose_name='name')),
                ('uuid', waldur_core.core.fields.UUIDField()),
                ('error_message', models.TextField(blank=True)),
                ('schedule', waldur_core.core.fields.CronScheduleField(max_length=15, validators=[waldur_core.core.validators.validate_cron_schedule, waldur_core.core.validators.MinCronValueValidator(1)])),
                ('next_trigger_at', models.DateTimeField(null=True)),
                ('timezone', models.CharField(default=django.utils.timezone.get_current_timezone_name, max_length=50)),
                ('is_active', models.BooleanField(default=False)),
                ('state', django_fsm.FSMIntegerField(choices=[(5, 'Creation Scheduled'), (6, 'Creating'), (1, 'Update Scheduled'), (2, 'Updating'), (7, 'Deletion Scheduled'), (8, 'Deleting'), (3, 'OK'), (4, 'Erred')], default=5)),
                ('backend_id', models.CharField(blank=True, max_length=255)),
                ('retention_time', models.PositiveIntegerField(help_text='Retention time in days, if 0 - resource will be kept forever')),
                ('maximal_number_of_resources', models.PositiveSmallIntegerField()),
                ('call_count', models.PositiveSmallIntegerField(default=0, help_text='How many times a resource schedule was called.')),
                ('service_project_link', models.ForeignKey(on_delete=django.db.models.deletion.PROTECT, related_name='snapshot_schedules', to='openstack_tenant.OpenStackTenantServiceProjectLink')),
            ],
            options={
                'abstract': False,
            },
            bases=(waldur_core.core.models.DescendantMixin, waldur_core.core.models.BackendModelMixin, waldur_core.structure.models.StructureLoggableMixin, models.Model),
        ),
        migrations.CreateModel(
            name='SubNet',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('description', models.CharField(blank=True, max_length=500, verbose_name='description')),
                ('name', models.CharField(max_length=150, validators=[waldur_core.core.validators.validate_name], verbose_name='name')),
                ('uuid', waldur_core.core.fields.UUIDField()),
                ('backend_id', models.CharField(db_index=True, max_length=255)),
                ('cidr', models.CharField(blank=True, max_length=32)),
                ('gateway_ip', models.GenericIPAddressField(null=True, protocol='IPv4')),
                ('allocation_pools', waldur_core.core.fields.JSONField(default=dict)),
                ('ip_version', models.SmallIntegerField(default=4)),
                ('enable_dhcp', models.BooleanField(default=True)),
                ('dns_nameservers', waldur_core.core.fields.JSONField(default=list, help_text='List of DNS name servers associated with the subnet.')),
                ('network', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='subnets', to='openstack_tenant.Network')),
                ('settings', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='+', to='structure.ServiceSettings')),
            ],
            options={
                'verbose_name': 'Subnet',
                'verbose_name_plural': 'Subnets',
            },
            bases=(waldur_core.core.models.BackendModelMixin, models.Model),
        ),
        migrations.CreateModel(
            name='Volume',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('created', model_utils.fields.AutoCreatedField(default=django.utils.timezone.now, editable=False, verbose_name='created')),
                ('modified', model_utils.fields.AutoLastModifiedField(default=django.utils.timezone.now, editable=False, verbose_name='modified')),
                ('description', models.CharField(blank=True, max_length=500, verbose_name='description')),
                ('name', models.CharField(max_length=150, validators=[waldur_core.core.validators.validate_name], verbose_name='name')),
                ('uuid', waldur_core.core.fields.UUIDField()),
                ('error_message', models.TextField(blank=True)),
                ('runtime_state', models.CharField(blank=True, max_length=150, verbose_name='runtime state')),
                ('state', django_fsm.FSMIntegerField(choices=[(5, 'Creation Scheduled'), (6, 'Creating'), (1, 'Update Scheduled'), (2, 'Updating'), (7, 'Deletion Scheduled'), (8, 'Deleting'), (3, 'OK'), (4, 'Erred')], default=5)),
                ('size', models.PositiveIntegerField(help_text='Size in MiB')),
                ('backend_id', models.CharField(blank=True, max_length=255, null=True)),
                ('device', models.CharField(blank=True, help_text='Name of volume as instance device e.g. /dev/vdb.', max_length=50, validators=[django.core.validators.RegexValidator('^/dev/[a-zA-Z0-9]+$', message='Device should match pattern "/dev/alphanumeric+"')])),
                ('bootable', models.BooleanField(default=False)),
                ('metadata', waldur_core.core.fields.JSONField(blank=True)),
                ('image_name', models.CharField(blank=True, max_length=150)),
                ('image_metadata', waldur_core.core.fields.JSONField(blank=True)),
                ('action', models.CharField(blank=True, max_length=50)),
                ('action_details', waldur_core.core.fields.JSONField(default=dict)),
            ],
            bases=(waldur_openstack.openstack_tenant.models.TenantQuotaMixin, waldur_core.core.models.DescendantMixin, waldur_core.core.models.BackendModelMixin, waldur_core.structure.models.StructureLoggableMixin, models.Model),
        ),
        migrations.CreateModel(
            name='VolumeAvailabilityZone',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('name', models.CharField(max_length=150, validators=[waldur_core.core.validators.validate_name], verbose_name='name')),
                ('uuid', waldur_core.core.fields.UUIDField()),
                ('available', models.BooleanField(default=True)),
                ('settings', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='+', to='structure.ServiceSettings')),
            ],
            bases=(waldur_core.core.models.BackendModelMixin, models.Model),
        ),
        migrations.CreateModel(
            name='VolumeType',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('description', models.CharField(blank=True, max_length=500, verbose_name='description')),
                ('name', models.CharField(max_length=150, validators=[waldur_core.core.validators.validate_name], verbose_name='name')),
                ('uuid', waldur_core.core.fields.UUIDField()),
                ('backend_id', models.CharField(db_index=True, max_length=255)),
                ('settings', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='+', to='structure.ServiceSettings')),
            ],
            bases=(waldur_core.core.models.BackendModelMixin, models.Model),
        ),
        migrations.AddField(
            model_name='volume',
            name='availability_zone',
            field=models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.SET_NULL, to='openstack_tenant.VolumeAvailabilityZone'),
        ),
        migrations.AddField(
            model_name='volume',
            name='image',
            field=models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.SET_NULL, to='openstack_tenant.Image'),
        ),
        migrations.AddField(
            model_name='volume',
            name='instance',
            field=models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.CASCADE, related_name='volumes', to='openstack_tenant.Instance'),
        ),
        migrations.AddField(
            model_name='volume',
            name='service_project_link',
            field=models.ForeignKey(on_delete=django.db.models.deletion.PROTECT, related_name='volumes', to='openstack_tenant.OpenStackTenantServiceProjectLink'),
        ),
        migrations.AddField(
            model_name='volume',
            name='source_snapshot',
            field=models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.SET_NULL, related_name='volumes', to='openstack_tenant.Snapshot'),
        ),
        migrations.AddField(
            model_name='volume',
            name='tags',
            field=taggit.managers.TaggableManager(blank=True, help_text='A comma-separated list of tags.', through='taggit.TaggedItem', to='taggit.Tag', verbose_name='Tags'),
        ),
        migrations.AddField(
            model_name='volume',
            name='type',
            field=models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.SET_NULL, to='openstack_tenant.VolumeType'),
        ),
        migrations.AddField(
            model_name='snapshotschedule',
            name='source_volume',
            field=models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='snapshot_schedules', to='openstack_tenant.Volume'),
        ),
        migrations.AddField(
            model_name='snapshotschedule',
            name='tags',
            field=taggit.managers.TaggableManager(blank=True, help_text='A comma-separated list of tags.', through='taggit.TaggedItem', to='taggit.Tag', verbose_name='Tags'),
        ),
        migrations.AddField(
            model_name='snapshotrestoration',
            name='volume',
            field=models.OneToOneField(on_delete=django.db.models.deletion.CASCADE, related_name='restoration', to='openstack_tenant.Volume'),
        ),
        migrations.AddField(
            model_name='snapshot',
            name='snapshot_schedule',
            field=models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.SET_NULL, related_name='snapshots', to='openstack_tenant.SnapshotSchedule'),
        ),
        migrations.AddField(
            model_name='snapshot',
            name='source_volume',
            field=models.ForeignKey(null=True, on_delete=django.db.models.deletion.PROTECT, related_name='snapshots', to='openstack_tenant.Volume'),
        ),
        migrations.AddField(
            model_name='snapshot',
            name='tags',
            field=taggit.managers.TaggableManager(blank=True, help_text='A comma-separated list of tags.', through='taggit.TaggedItem', to='taggit.Tag', verbose_name='Tags'),
        ),
        migrations.AddField(
            model_name='openstacktenantservice',
            name='projects',
            field=models.ManyToManyField(related_name='openstack_tenant_services', through='openstack_tenant.OpenStackTenantServiceProjectLink', to='structure.Project'),
        ),
        migrations.AddField(
            model_name='openstacktenantservice',
            name='settings',
            field=models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='structure.ServiceSettings'),
        ),
        migrations.AddField(
            model_name='internalip',
            name='subnet',
            field=models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='internal_ips', to='openstack_tenant.SubNet'),
        ),
        migrations.AddField(
            model_name='instance',
            name='availability_zone',
            field=models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.SET_NULL, to='openstack_tenant.InstanceAvailabilityZone'),
        ),
        migrations.AddField(
            model_name='instance',
            name='security_groups',
            field=models.ManyToManyField(related_name='instances', to='openstack_tenant.SecurityGroup'),
        ),
        migrations.AddField(
            model_name='instance',
            name='service_project_link',
            field=models.ForeignKey(on_delete=django.db.models.deletion.PROTECT, related_name='instances', to='openstack_tenant.OpenStackTenantServiceProjectLink'),
        ),
        migrations.AddField(
            model_name='instance',
            name='subnets',
            field=models.ManyToManyField(through='openstack_tenant.InternalIP', to='openstack_tenant.SubNet'),
        ),
        migrations.AddField(
            model_name='instance',
            name='tags',
            field=taggit.managers.TaggableManager(blank=True, help_text='A comma-separated list of tags.', through='taggit.TaggedItem', to='taggit.Tag', verbose_name='Tags'),
        ),
        migrations.AddField(
            model_name='floatingip',
            name='internal_ip',
            field=models.ForeignKey(null=True, on_delete=django.db.models.deletion.SET_NULL, related_name='floating_ips', to='openstack_tenant.InternalIP'),
        ),
        migrations.AddField(
            model_name='floatingip',
            name='settings',
            field=models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='+', to='structure.ServiceSettings'),
        ),
        migrations.AddField(
            model_name='backupschedule',
            name='instance',
            field=models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='backup_schedules', to='openstack_tenant.Instance'),
        ),
        migrations.AddField(
            model_name='backupschedule',
            name='service_project_link',
            field=models.ForeignKey(on_delete=django.db.models.deletion.PROTECT, related_name='backup_schedules', to='openstack_tenant.OpenStackTenantServiceProjectLink'),
        ),
        migrations.AddField(
            model_name='backupschedule',
            name='tags',
            field=taggit.managers.TaggableManager(blank=True, help_text='A comma-separated list of tags.', through='taggit.TaggedItem', to='taggit.Tag', verbose_name='Tags'),
        ),
        migrations.AddField(
            model_name='backuprestoration',
            name='flavor',
            field=models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.SET_NULL, related_name='+', to='openstack_tenant.Flavor'),
        ),
        migrations.AddField(
            model_name='backuprestoration',
            name='instance',
            field=models.OneToOneField(on_delete=django.db.models.deletion.CASCADE, related_name='+', to='openstack_tenant.Instance'),
        ),
        migrations.AddField(
            model_name='backup',
            name='backup_schedule',
            field=models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.SET_NULL, related_name='backups', to='openstack_tenant.BackupSchedule'),
        ),
        migrations.AddField(
            model_name='backup',
            name='instance',
            field=models.ForeignKey(on_delete=django.db.models.deletion.PROTECT, related_name='backups', to='openstack_tenant.Instance'),
        ),
        migrations.AddField(
            model_name='backup',
            name='service_project_link',
            field=models.ForeignKey(on_delete=django.db.models.deletion.PROTECT, related_name='backups', to='openstack_tenant.OpenStackTenantServiceProjectLink'),
        ),
        migrations.AddField(
            model_name='backup',
            name='snapshots',
            field=models.ManyToManyField(related_name='backups', to='openstack_tenant.Snapshot'),
        ),
        migrations.AddField(
            model_name='backup',
            name='tags',
            field=taggit.managers.TaggableManager(blank=True, help_text='A comma-separated list of tags.', through='taggit.TaggedItem', to='taggit.Tag', verbose_name='Tags'),
        ),
        migrations.AlterUniqueTogether(
            name='volumetype',
            unique_together=set([('settings', 'backend_id')]),
        ),
        migrations.AlterUniqueTogether(
            name='volumeavailabilityzone',
            unique_together=set([('settings', 'name')]),
        ),
        migrations.AlterUniqueTogether(
            name='volume',
            unique_together=set([('service_project_link', 'backend_id')]),
        ),
        migrations.AlterUniqueTogether(
            name='subnet',
            unique_together=set([('settings', 'backend_id')]),
        ),
        migrations.AlterUniqueTogether(
            name='snapshot',
            unique_together=set([('service_project_link', 'backend_id')]),
        ),
        migrations.AlterUniqueTogether(
            name='securitygroup',
            unique_together=set([('settings', 'backend_id')]),
        ),
        migrations.AlterUniqueTogether(
            name='openstacktenantserviceprojectlink',
            unique_together=set([('service', 'project')]),
        ),
        migrations.AlterUniqueTogether(
            name='openstacktenantservice',
            unique_together=set([('customer', 'settings')]),
        ),
        migrations.AlterUniqueTogether(
            name='network',
            unique_together=set([('settings', 'backend_id')]),
        ),
        migrations.AlterUniqueTogether(
            name='internalip',
            unique_together=set([('backend_id', 'settings')]),
        ),
        migrations.AlterUniqueTogether(
            name='instanceavailabilityzone',
            unique_together=set([('settings', 'name')]),
        ),
        migrations.AlterUniqueTogether(
            name='instance',
            unique_together=set([('service_project_link', 'backend_id')]),
        ),
        migrations.AlterUniqueTogether(
            name='image',
            unique_together=set([('settings', 'backend_id')]),
        ),
        migrations.AlterUniqueTogether(
            name='floatingip',
            unique_together=set([('settings', 'address')]),
        ),
        migrations.AlterUniqueTogether(
            name='flavor',
            unique_together=set([('settings', 'backend_id')]),
        ),
    ]



from celery import chain

from waldur_core.core import executors as core_executors
from waldur_core.core import tasks as core_tasks
from waldur_core.core import utils as core_utils

from . import models, signals


def pull_datastores_for_resource(instance, task):
    """
    Schedule datastore synchronization after virtual machine or disk
    has been either created, updated or deleted.
    """

    if isinstance(instance, models.VirtualMachine):
        datastore = instance.datastore
    elif isinstance(instance, models.Disk):
        datastore = instance.vm.datastore
    else:
        datastore = None

    if not datastore:
        return task

    serialized_settings = core_utils.serialize_instance(instance.service_settings)
    return chain(task, core_tasks.IndependentBackendMethodTask().si(
        serialized_settings,
        'pull_datastores',
    ))


class VirtualMachinePullExecutor(core_executors.ActionExecutor):
    action = 'Pull'

    @classmethod
    def get_task_signature(cls, instance, serialized_instance, **kwargs):
        return core_tasks.BackendMethodTask().si(
            serialized_instance,
            'pull_virtual_machine',
            state_transition='begin_updating'
        )


class VirtualMachineCreateExecutor(core_executors.CreateExecutor):

    @classmethod
    def get_task_signature(cls, instance, serialized_instance, **kwargs):
        task = core_tasks.BackendMethodTask().si(
            serialized_instance,
            'create_virtual_machine',
            state_transition='begin_creating'
        )
        return chain(
            pull_datastores_for_resource(instance, task),
            core_tasks.BackendMethodTask().si(
                serialized_instance,
                'pull_vm_ports',
            ),
            core_tasks.BackendMethodTask().si(
                serialized_instance,
                'pull_virtual_machine',
            ),
        )


class VirtualMachineDeleteExecutor(core_executors.DeleteExecutor):

    @classmethod
    def get_task_signature(cls, instance, serialized_instance, **kwargs):
        if instance.backend_id:
            task = core_tasks.BackendMethodTask().si(
                serialized_instance,
                'delete_virtual_machine',
                state_transition='begin_deleting')
            return pull_datastores_for_resource(instance, task)
        else:
            return core_tasks.StateTransitionTask().si(
                serialized_instance,
                state_transition='begin_deleting'
            )


class VirtualMachineStartExecutor(core_executors.ActionExecutor):
    action = 'Start'

    @classmethod
    def get_task_signature(cls, instance, serialized_instance, **kwargs):
        _tasks = [
            core_tasks.BackendMethodTask().si(
                serialized_instance,
                'start_virtual_machine',
                state_transition='begin_updating'
            )
        ]
        if instance.tools_installed:
            _tasks.append(
                core_tasks.BackendMethodTask().si(
                    serialized_instance,
                    'pull_virtual_machine',
                )
            )
            _tasks.append(
                core_tasks.PollBackendCheckTask().si(
                    serialized_instance,
                    'is_virtual_machine_tools_running'
                )
            )
        _tasks.append(
            core_tasks.BackendMethodTask().si(
                serialized_instance,
                'pull_virtual_machine',
            )
        )
        return chain(_tasks)


class VirtualMachineStopExecutor(core_executors.ActionExecutor):
    action = 'Stop'

    @classmethod
    def get_task_signature(cls, instance, serialized_instance, **kwargs):
        return chain(
            core_tasks.BackendMethodTask().si(
                serialized_instance,
                'stop_virtual_machine',
                state_transition='begin_updating'
            ),
            core_tasks.BackendMethodTask().si(
                serialized_instance,
                'pull_virtual_machine',
            ),
        )


class VirtualMachineResetExecutor(core_executors.ActionExecutor):
    action = 'Reset'

    @classmethod
    def pre_apply(cls, instance, **kwargs):
        super(VirtualMachineResetExecutor, cls).pre_apply(instance, **kwargs)
        instance.tools_state = models.VirtualMachine.ToolsStates.NOT_RUNNING
        instance.save(update_fields=['tools_state'])

    @classmethod
    def get_task_signature(cls, instance, serialized_instance, **kwargs):
        _tasks = [
            core_tasks.BackendMethodTask().si(
                serialized_instance,
                'reset_virtual_machine',
                state_transition='begin_updating'
            )
        ]
        if instance.tools_installed:
            _tasks.append(
                core_tasks.PollBackendCheckTask().si(
                    serialized_instance,
                    'is_virtual_machine_tools_not_running'
                )
            )
            _tasks.append(
                core_tasks.PollBackendCheckTask().si(
                    serialized_instance,
                    'is_virtual_machine_tools_running'
                )
            )
        _tasks.append(
            core_tasks.BackendMethodTask().si(
                serialized_instance,
                'pull_virtual_machine',
            )
        )
        return chain(_tasks)


class VirtualMachineSuspendExecutor(core_executors.ActionExecutor):
    action = 'Suspend'

    @classmethod
    def get_task_signature(cls, instance, serialized_instance, **kwargs):
        return chain(
            core_tasks.BackendMethodTask().si(
                serialized_instance,
                'suspend_virtual_machine',
                state_transition='begin_updating'
            ),
            core_tasks.BackendMethodTask().si(
                serialized_instance,
                'pull_virtual_machine',
            ),
        )


class VirtualMachineShutdownGuestExecutor(core_executors.ActionExecutor):
    action = 'Shutdown Guest'

    @classmethod
    def pre_apply(cls, instance, **kwargs):
        super(VirtualMachineShutdownGuestExecutor, cls).pre_apply(instance, **kwargs)
        instance.tools_state = models.VirtualMachine.ToolsStates.NOT_RUNNING
        instance.save(update_fields=['tools_state'])

    @classmethod
    def get_task_signature(cls, instance, serialized_instance, **kwargs):
        return chain(
            core_tasks.BackendMethodTask().si(
                serialized_instance,
                'shutdown_guest',
                state_transition='begin_updating'
            ),
            core_tasks.PollBackendCheckTask().si(
                serialized_instance,
                'is_virtual_machine_shutted_down'
            ),
            core_tasks.BackendMethodTask().si(
                serialized_instance,
                'pull_virtual_machine',
            ),
        )


class VirtualMachineRebootGuestExecutor(core_executors.ActionExecutor):
    action = 'Reboot Guest'

    @classmethod
    def pre_apply(cls, instance, **kwargs):
        super(VirtualMachineRebootGuestExecutor, cls).pre_apply(instance, **kwargs)
        instance.tools_state = models.VirtualMachine.ToolsStates.NOT_RUNNING
        instance.save(update_fields=['tools_state'])

    @classmethod
    def get_task_signature(cls, instance, serialized_instance, **kwargs):
        return chain(
            core_tasks.BackendMethodTask().si(
                serialized_instance,
                'reboot_guest',
                state_transition='begin_updating'
            ),
            core_tasks.PollBackendCheckTask().si(
                serialized_instance,
                'is_virtual_machine_tools_not_running'
            ),
            core_tasks.PollBackendCheckTask().si(
                serialized_instance,
                'is_virtual_machine_tools_running'
            ),
            core_tasks.BackendMethodTask().si(
                serialized_instance,
                'pull_virtual_machine',
            ),
        )


class VirtualMachineUpdateExecutor(core_executors.UpdateExecutor):

    @classmethod
    def get_task_signature(cls, instance, serialized_instance, **kwargs):
        if {'cores', 'cores_per_socket', 'ram'} & set(kwargs['updated_fields']):
            return core_tasks.BackendMethodTask().si(
                serialized_instance,
                'update_virtual_machine',
                state_transition='begin_updating'
            )
        else:
            return core_tasks.StateTransitionTask().si(
                serialized_instance,
                state_transition='begin_updating'
            )


class PortCreateExecutor(core_executors.CreateExecutor):

    @classmethod
    def get_task_signature(cls, instance, serialized_instance, **kwargs):
        task = core_tasks.BackendMethodTask().si(
            serialized_instance,
            'create_port',
            state_transition='begin_creating'
        )
        return chain(
            task,
            core_tasks.BackendMethodTask().si(
                serialized_instance,
                'pull_port',
            )
        )


class PortPullExecutor(core_executors.ActionExecutor):
    action = 'Pull'

    @classmethod
    def get_task_signature(cls, instance, serialized_instance, **kwargs):
        return core_tasks.BackendMethodTask().si(
            serialized_instance,
            'pull_port',
            state_transition='begin_updating'
        )


class PortDeleteExecutor(core_executors.DeleteExecutor):

    @classmethod
    def get_task_signature(cls, instance, serialized_instance, **kwargs):
        if instance.backend_id:
            return core_tasks.BackendMethodTask().si(
                serialized_instance,
                'delete_port',
                state_transition='begin_deleting'
            )
        else:
            return core_tasks.StateTransitionTask().si(
                serialized_instance,
                state_transition='begin_deleting'
            )


class DiskPullExecutor(core_executors.ActionExecutor):
    action = 'Pull'

    @classmethod
    def get_task_signature(cls, instance, serialized_instance, **kwargs):
        return core_tasks.BackendMethodTask().si(
            serialized_instance,
            'pull_disk',
            state_transition='begin_updating'
        )


class DiskCreateExecutor(core_executors.CreateExecutor):

    @classmethod
    def get_task_signature(cls, instance, serialized_instance, **kwargs):
        task = core_tasks.BackendMethodTask().si(
            serialized_instance,
            'create_disk',
            state_transition='begin_creating'
        )
        task = pull_datastores_for_resource(instance, task)
        return chain(
            task,
            core_tasks.BackendMethodTask().si(
                serialized_instance,
                'pull_disk',
            )
        )


class DiskDeleteExecutor(core_executors.DeleteExecutor):

    @classmethod
    def get_task_signature(cls, instance, serialized_instance, **kwargs):
        if instance.backend_id:
            task = core_tasks.BackendMethodTask().si(
                serialized_instance,
                'delete_disk',
                state_transition='begin_deleting'
            )
            return pull_datastores_for_resource(instance, task)
        else:
            return core_tasks.StateTransitionTask().si(
                serialized_instance,
                state_transition='begin_deleting'
            )

    @classmethod
    def get_deletion_task(cls, instance, serialized_instance):
        serialized_vm = core_utils.serialize_instance(instance.vm)
        return chain(
            core_tasks.DeletionTask().si(serialized_instance),
            VirtualMachineUpdatedNotificationTask().si(serialized_vm),
        )

    @classmethod
    def get_success_signature(cls, instance, serialized_instance, **kwargs):
        return cls.get_deletion_task(instance, serialized_instance)

    @classmethod
    def get_failure_signature(cls, instance, serialized_instance, force=False, **kwargs):
        if force:
            return cls.get_deletion_task(instance, serialized_instance)
        else:
            return core_tasks.ErrorStateTransitionTask().s(serialized_instance)


class VirtualMachineUpdatedNotificationTask(core_tasks.Task):
    """
    Send notification that virtual machine configuration has been changed on backend
    so that it would be possible for a signal listener to invalidate their internal cache.
    It is needed, for example, after virtual disk is deleted from both VMware and Waldur
    because resource configuration in the marketplace is obtained from local database cache.
    """

    def execute(self, instance):
        signals.vm_updated.send(self.__class__, vm=instance)


class DiskExtendExecutor(core_executors.ActionExecutor):
    action = 'Extend'

    @classmethod
    def get_task_signature(cls, instance, serialized_instance, **kwargs):
        task = core_tasks.BackendMethodTask().si(
            serialized_instance, 'extend_disk',
            state_transition='begin_updating')
        return pull_datastores_for_resource(instance, task)



# -*- coding: utf-8 -*-
# Generated by Django 1.11.21 on 2019-08-29 08:49
from __future__ import unicode_literals

import django.core.validators
from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('marketplace', '0091_offering_paused_reason'),
    ]

    operations = [
        migrations.AlterField(
            model_name='attribute',
            name='key',
            field=models.CharField(max_length=255, primary_key=True, serialize=False, validators=[django.core.validators.RegexValidator('^[a-zA-Z][a-zA-Z0-9_]+$')]),
        ),
        migrations.AlterField(
            model_name='attributeoption',
            name='key',
            field=models.CharField(max_length=255, validators=[django.core.validators.RegexValidator('^[a-zA-Z][a-zA-Z0-9_]+$')]),
        ),
        migrations.AlterField(
            model_name='categorycomponent',
            name='type',
            field=models.CharField(help_text='Unique internal name of the measured unit, for example floating_ip.', max_length=50, validators=[django.core.validators.RegexValidator('^[a-zA-Z][a-zA-Z0-9_]+$')]),
        ),
        migrations.AlterField(
            model_name='offeringcomponent',
            name='type',
            field=models.CharField(help_text='Unique internal name of the measured unit, for example floating_ip.', max_length=50, validators=[django.core.validators.RegexValidator('^[a-zA-Z][a-zA-Z0-9_]+$')]),
        ),
    ]



# -*- coding: utf-8 -*-
# Generated by Django 1.11.13 on 2018-08-24 08:37
from __future__ import unicode_literals

import django.contrib.postgres.fields.jsonb
from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ('marketplace', '0025_private_offering'),
    ]

    operations = [
        migrations.AddField(
            model_name='offering',
            name='options',
            field=django.contrib.postgres.fields.jsonb.JSONField(blank=True, default=dict,
                                                                 help_text='Fields describing Offering request form.'),
        ),
    ]



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from django.db import migrations, models


def copy_external_ips(apps, schema_editor):
    Instance = apps.get_model('openstack_tenant', 'Instance')
    FloatingIP = apps.get_model('openstack_tenant', 'FloatingIP')

    for instance in Instance.objects.iterator():
        floating_ip = FloatingIP.objects.filter(
            settings=instance.service_project_link.service.settings,
            address=instance.external_ips,
        ).first()

        if floating_ip:
            instance.external_ip = floating_ip
            instance.save()


class Migration(migrations.Migration):

    dependencies = [
        ('openstack_tenant', '0018_remove_openstacktenantservice_name'),
    ]

    operations = [
        migrations.RemoveField(
            model_name='instance',
            name='internal_ips',
        ),
        migrations.AddField(
            model_name='instance',
            name='external_ip',
            field=models.OneToOneField(related_name='instance', to='openstack_tenant.FloatingIP', blank=True, null=True, on_delete=models.SET_NULL),
        ),
        migrations.RunPython(copy_external_ips, elidable=True),
    ]



# -*- coding: utf-8 -*-
# Generated by Django 1.11.18 on 2019-03-18 09:30
from __future__ import unicode_literals

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('marketplace', '0065_add_resource_cost_estimate'),
    ]

    operations = [
        migrations.AddField(
            model_name='cartitem',
            name='cost',
            field=models.DecimalField(blank=True, decimal_places=10, max_digits=22, null=True),
        ),
    ]



# -*- coding: utf-8 -*-
# Generated by Django 1.11.1 on 2017-06-20 16:16
from __future__ import unicode_literals

from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    dependencies = [
        ('structure', '0052_customer_subnets'),
        ('invoices', '0011_add_article_code'),
    ]

    operations = [
        migrations.AddField(
            model_name='offeringitem',
            name='project',
            field=models.ForeignKey(null=True, on_delete=django.db.models.deletion.SET_NULL, to='structure.Project'),
        ),
        migrations.AddField(
            model_name='offeringitem',
            name='project_name',
            field=models.CharField(blank=True, max_length=150),
        ),
        migrations.AddField(
            model_name='offeringitem',
            name='project_uuid',
            field=models.CharField(blank=True, max_length=32),
        ),
        migrations.AddField(
            model_name='openstackitem',
            name='project',
            field=models.ForeignKey(null=True, on_delete=django.db.models.deletion.SET_NULL, to='structure.Project'),
        ),
        migrations.AddField(
            model_name='openstackitem',
            name='project_name',
            field=models.CharField(blank=True, max_length=150),
        ),
        migrations.AddField(
            model_name='openstackitem',
            name='project_uuid',
            field=models.CharField(blank=True, max_length=32),
        ),
    ]



# -*- coding: utf-8 -*-
# Generated by Django 1.11.21 on 2019-09-09 13:20
from __future__ import unicode_literals

from decimal import Decimal
import django.contrib.postgres.fields.jsonb
import django.core.validators
from django.db import migrations, models
import django.db.models.deletion
import django.utils.timezone
import waldur_core.core.fields
import waldur_mastermind.invoices.models
import waldur_mastermind.invoices.utils


class Migration(migrations.Migration):

    replaces = [(b'invoices', '0001_initial'), (b'invoices', '0002_invoice_additional_details'), (b'invoices', '0003_paymentdetails_default_tax_percent'), (b'invoices', '0004_paymentdetails_uuid'), (b'invoices', '0005_add_company_type'), (b'invoices', '0006_add_daily_price'), (b'invoices', '0007_item_remove_price'), (b'invoices', '0008_offeringitem'), (b'invoices', '0009_quantize_prices'), (b'invoices', '0010_add_product_code'), (b'invoices', '0011_add_article_code'), (b'invoices', '0012_add_project_to_invoice_item'), (b'invoices', '0013_set_projects_for_invoice_items'), (b'invoices', '0014_add_accounting_start_date'), (b'invoices', '0015_add_unit_price'), (b'invoices', '0016_remove_daily_price'), (b'invoices', '0017_genericinvoiceitem'), (b'invoices', '0018_genericinvoiceitem_quantity'), (b'invoices', '0019_mark_payments_details_fields'), (b'invoices', '0020_migrate_payment_details_data'), (b'invoices', '0021_remove_pending_offering_item'), (b'invoices', '0022_remove_payment_details'), (b'invoices', '0023_invoice_current_cost'), (b'invoices', '0024_immutable_default_json'), (b'invoices', '0025_servicedowntime'), (b'invoices', '0026_invoice__file'), (b'invoices', '0027_migrate_items'), (b'invoices', '0028_delete_offering_and_openstack_item'), (b'invoices', '0029_migrate_invoice_item_name'), (b'invoices', '0030_json_details_on_genericinvoiceitem')]

    initial = True

    dependencies = [
        ('contenttypes', '0002_remove_content_type_name'),
        ('support', '0010_add_unit_price'),
        ('packages', '0002_openstack_packages'),
        ('packages', '0013_packagetemplate_article_code'),
        ('structure', '0054_payment_details'),
        ('structure', '0037_remove_customer_billing_backend_id'),
        ('support', '0006_offering'),
        ('packages', '0014_unique_package_tenant'),
        ('support', '0011_remove_price'),
        ('support', '0009_offering_article_code'),
        ('structure', '0052_customer_subnets'),
    ]

    operations = [
        migrations.CreateModel(
            name='Invoice',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('uuid', waldur_core.core.fields.UUIDField()),
                ('month', models.PositiveSmallIntegerField(default=waldur_mastermind.invoices.utils.get_current_month, validators=[django.core.validators.MinValueValidator(1), django.core.validators.MaxValueValidator(12)])),
                ('year', models.PositiveSmallIntegerField(default=waldur_mastermind.invoices.utils.get_current_year)),
                ('state', models.CharField(choices=[('pending', 'Pending'), ('created', 'Created'), ('paid', 'Paid'), ('canceled', 'Canceled')], default='pending', max_length=30)),
                ('customer', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='+', to='structure.Customer', verbose_name='organization')),
                ('invoice_date', models.DateField(blank=True, help_text='Date then invoice moved from state pending to created.', null=True)),
                ('tax_percent', models.DecimalField(decimal_places=2, default=0, max_digits=4, validators=[django.core.validators.MinValueValidator(0), django.core.validators.MaxValueValidator(100)])),
            ],
        ),
        migrations.AlterUniqueTogether(
            name='invoice',
            unique_together=set([('customer', 'month', 'year')]),
        ),
        migrations.CreateModel(
            name='GenericInvoiceItem',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('unit_price', models.DecimalField(decimal_places=7, default=0, max_digits=22, validators=[django.core.validators.MinValueValidator(Decimal('0'))])),
                ('unit', models.CharField(choices=[(b'month', 'Per month'), (b'half_month', 'Per half month'), (b'day', 'Per day'), (b'hour', 'Per hour'), (b'quantity', 'Quantity')], default=b'day', max_length=30)),
                ('product_code', models.CharField(blank=True, max_length=30)),
                ('article_code', models.CharField(blank=True, max_length=30)),
                ('start', models.DateTimeField(default=waldur_mastermind.invoices.utils.get_current_month_start, help_text='Date and time when item usage has started.')),
                ('end', models.DateTimeField(default=waldur_mastermind.invoices.utils.get_current_month_end, help_text='Date and time when item usage has ended.')),
                ('project_name', models.CharField(blank=True, max_length=150)),
                ('project_uuid', models.CharField(blank=True, max_length=32)),
                ('object_id', models.PositiveIntegerField(null=True)),
                ('details', django.contrib.postgres.fields.jsonb.JSONField(blank=True, default=dict, help_text='Stores data about scope')),
                ('content_type', models.ForeignKey(null=True, on_delete=django.db.models.deletion.CASCADE, related_name='+', to='contenttypes.ContentType')),
                ('invoice', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='generic_items', to='invoices.Invoice')),
                ('project', models.ForeignKey(null=True, on_delete=django.db.models.deletion.SET_NULL, to='structure.Project')),
                ('quantity', models.PositiveIntegerField(default=0)),
            ],
            options={
                'abstract': False,
            },
        ),
        migrations.AddField(
            model_name='invoice',
            name='current_cost',
            field=models.DecimalField(decimal_places=2, default=0, editable=False, help_text='Cached value for current cost.', max_digits=10),
        ),
        migrations.CreateModel(
            name='ServiceDowntime',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('start', models.DateTimeField(default=waldur_mastermind.invoices.models.get_default_downtime_start, help_text='Date and time when downtime has started.')),
                ('end', models.DateTimeField(default=django.utils.timezone.now, help_text='Date and time when downtime has ended.')),
                ('package', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='packages.OpenStackPackage')),
            ],
        ),
        migrations.AddField(
            model_name='invoice',
            name='_file',
            field=models.TextField(blank=True, editable=False),
        ),
    ]
